###############################################################################

# Reader for CINE files produced by Vision Research Phantom Software
# Author: Dustin Kleckner
# dkleckner@uchicago.edu

# Modified by Thomas A Caswell (tcaswell@uchicago.edu)
# Added to PIMS by Thomas A Caswell (tcaswell@gmail.com)

# Modified by B. Neel
###############################################################################

from pims.frame import Frame
from pims.base_frames import FramesSequence, index_attr
from pims.utils.misc import FileLocker
import time
import struct
import numpy as np
from numpy import array, frombuffer, where
from threading import Lock
import datetime
import hashlib
import warnings
from collections.abc import Iterable

__all__ = ('Cine', )


# '<' for little endian (cine documentation)
def _build_struct(dtype):
    return struct.Struct(str("<" + dtype))


FRACTION_MASK = (2**32-1)
MAX_INT = 2**32

# Harmonized/simplified cine file data types with Python struct doc
UINT8 = 'B'
CHAR = 'b'
UINT16 = 'H'
INT16 = 'h'
BOOL = 'i'
UINT32 = 'I'
INT32 = 'i'
INT64 = 'q'
FLOAT = 'f'
DOUBLE = 'd'
TIME64 = 'Q'
RECT = '4i'
WBGAIN = '2f'
IMFILTER = '28i'
# TODO: get correct format for TrigTC
TC = '8s'

CFA_NONE = 0    # gray sensor
CFA_VRI = 1     # gbrg/rggb sensor
CFA_VRIV6 = 2   # bggr/grbg sensor
CFA_BAYER = 3   # gb/rg sensor
CFA_BAYERFLIP = 4   #rg/gb sensor

TAGGED_FIELDS = {
    1000: ('ang_dig_sigs', ''),
    1001: ('image_time_total', TIME64),
    1002: ('image_time_only', TIME64),
    1003: ('exposure_only', UINT32),
    1004: ('range_data', ''),
    1005: ('binsig', ''),
    1006: ('anasig', ''),
    1007: ('time_code', '')}

HEADER_FIELDS = [
    ('type', '2s'),
    ('header_size', UINT16),
    ('compression', UINT16),
    ('version', UINT16),
    ('first_movie_image', INT32),
    ('total_image_count', UINT32),
    ('first_image_no', INT32),
    ('image_count', UINT32),
    # Offsets of following sections
    ('off_image_header', UINT32),
    ('off_setup', UINT32),
    ('off_image_offsets', UINT32),
    ('trigger_time', TIME64),
]

BITMAP_INFO_FIELDS = [
    ('bi_size', UINT32),
    ('bi_width', INT32),
    ('bi_height', INT32),
    ('bi_planes', UINT16),
    ('bi_bit_count', UINT16),
    ('bi_compression', UINT32),
    ('bi_image_size', UINT32),
    ('bi_x_pels_per_meter', INT32),
    ('bi_y_pels_per_meter', INT32),
    ('bi_clr_used', UINT32),
    ('bi_clr_important', UINT32),
]


SETUP_FIELDS = [
    ('frame_rate_16', UINT16),
    ('shutter_16', UINT16),
    ('post_trigger_16', UINT16),
    ('frame_delay_16', UINT16),
    ('aspect_ratio', UINT16),
    ('contrast_16', UINT16),
    ('bright_16', UINT16),
    ('rotate_16', UINT8),
    ('time_annotation', UINT8),
    ('trig_cine', UINT8),
    ('trig_frame', UINT8),
    ('shutter_on', UINT8),
    ('description_old', '121s'),
    ('mark', '2s'),
    ('length', UINT16),
    ('binning', UINT16),
    ('sig_option', UINT16),
    ('bin_channels', INT16),
    ('samples_per_image', UINT8),
    ] + [('bin_name{:d}'.format(i), '11s') for i in range(8)] + [
    ('ana_option', UINT16),
    ('ana_channels', INT16),
    ('res_6', UINT8),
    ('ana_board', UINT8),
    ] + [('ch_option{:d}'.format(i), INT16) for i in range(8)] + [
    ] + [('ana_gain{:d}'.format(i), FLOAT) for i in range(8)] + [
    ] + [('ana_unit{:d}'.format(i), '6s') for i in range(8)] + [
    ] + [('ana_name{:d}'.format(i), '11s') for i in range(8)] + [
    ('i_first_image', INT32),
    ('dw_image_count', UINT32),
    ('n_q_factor', INT16),
    ('w_cine_file_type', UINT16),
    ] + [('sz_cine_path{:d}'.format(i), '65s') for i in range(4)] + [
    ('b_mains_freq', UINT16),
    ('b_time_code', UINT8),
    ('b_priority', UINT8),
    ('w_leap_sec_dy', UINT16),
    ('d_delay_tc', DOUBLE),
    ('d_delay_pps', DOUBLE),
    ('gen_bits', UINT16),
    ('res_1', INT32),  
    ('res_2', INT32),
    ('res_3', INT32),
    ('im_width', UINT16),
    ('im_height', UINT16),
    ('edr_shutter_16', UINT16),
    ('serial', UINT32),
    ('saturation', INT32),
    ('res_5', UINT8),
    ('auto_exposure', UINT32),
    ('b_flip_h', BOOL),
    ('b_flip_v', BOOL),
    ('grid', UINT32),
    ('frame_rate', UINT32),
    ('shutter', UINT32),
    ('edr_shutter', UINT32),
    ('post_trigger', UINT32),
    ('frame_delay', UINT32),
    ('b_enable_color', BOOL),
    ('camera_version', UINT32),
    ('firmware_version', UINT32),
    ('software_version', UINT32),
    ('recording_time_zone', INT32),
    ('cfa', UINT32),
    ('bright', INT32),
    ('contrast', INT32),
    ('gamma', INT32),
    ('res_21', UINT32),
    ('auto_exp_level', UINT32),
    ('auto_exp_speed', UINT32),
    ('auto_exp_rect', RECT),
    ('wb_gain', '8f'),
    ('rotate', INT32),
    ('wb_view', WBGAIN),
    ('real_bpp', UINT32),
    ('conv_8_min', UINT32),
    ('conv_8_max', UINT32),
    ('filter_code', INT32),
    ('filter_param', INT32),
    ('uf', IMFILTER),
    ('black_cal_sver', UINT32),
    ('white_cal_sver', UINT32),
    ('gray_cal_sver', UINT32),
    ('b_stamp_time', BOOL),
    ('sound_dest', UINT32),
    ('frp_steps', UINT32),
    ] + [('frp_img_nr{:d}'.format(i), INT32) for i in range(16)] + [
    ] + [('frp_rate{:d}'.format(i), UINT32) for i in range(16)] + [
    ] + [('frp_exp{:d}'.format(i), UINT32) for i in range(16)] + [
    ('mc_cnt', INT32),
    ] + [('mc_percent{:d}'.format(i), FLOAT) for i in range(64)] + [
    ('ci_calib', UINT32),
    ('calib_width', UINT32),
    ('calib_height', UINT32),
    ('calib_rate', UINT32),
    ('calib_exp', UINT32),
    ('calib_edr', UINT32),
    ('calib_temp', UINT32),
    ] + [('header_serial{:d}'.format(i), UINT32) for i in range(4)] + [
    ('range_code', UINT32),
    ('range_size', UINT32),
    ('decimation', UINT32),
    ('master_serial', UINT32),
    ('sensor', UINT32),
    ('shutter_ns', UINT32),
    ('edr_shutter_ns', UINT32),
    ('frame_delay_ns', UINT32),
    ('im_pos_xacq', UINT32),
    ('im_pos_yacq', UINT32),
    ('im_width_acq', UINT32),
    ('im_height_acq', UINT32),
    ('description', '4096s'),
    ('rising_edge', BOOL),
    ('filter_time', UINT32),
    ('long_ready', BOOL),
    ('shutter_off', BOOL),
    ('res_4', '16s'),
    ('b_meta_WB', BOOL),
    ('hue', INT32),
    ('black_level', INT32),
    ('white_level', INT32),
    ('lens_description', '256s'),
    ('lens_aperture', FLOAT),
    ('lens_focus_distance', FLOAT),
    ('lens_focal_length', FLOAT),
    ('f_offset', FLOAT),
    ('f_gain', FLOAT),
    ('f_saturation', FLOAT),
    ('f_hue', FLOAT),
    ('f_gamma', FLOAT),
    ('f_gamma_R', FLOAT),
    ('f_gamma_B', FLOAT),
    ('f_flare', FLOAT),
    ('f_pedestal_R', FLOAT),
    ('f_pedestal_G', FLOAT),
    ('f_pedestal_B', FLOAT),
    ('f_chroma', FLOAT),
    ('tone_label', '256s'),
    ('tone_points', INT32),
    ('f_tone', ''.join(32*['2f'])),
    ('user_matrix_label', '256s'),
    ('enable_matrices', BOOL),
    ('f_user_matrix', '9'+FLOAT),
    ('enable_crop', BOOL),
    ('crop_left_top_right_bottom', '4i'),
    ('enable_resample', BOOL),
    ('resample_width', UINT32),
    ('resample_height', UINT32),
    ('f_gain16_8', FLOAT),
    ('frp_shape', '16'+UINT32),
    ('trig_TC', TC),
    ('f_pb_rate', FLOAT),
    ('f_tc_rate', FLOAT),
    ('cine_name', '256s')
]

#from VR doc: This field is maintained for compatibility with old versions but
#a new field was added for that information. The new field can be larger or may
#have a different measurement unit.
UPDATED_FIELDS = {
        'frame_rate_16': 'frame_rate',
        'shutter_16': 'shutter_ns',
        'post_trigger_16': 'post_trigger',
        'frame_delay_16': 'frame_delay_ns',
        'edr_shutter_16': 'edr_shutter_ns',
        'saturation': 'f_saturation',
        'shutter': 'shutter_ns',
        'edr_shutter': 'edr_shutter_ns',
        'frame_delay': 'frame_delay_ns',
        'bright': 'f_offset',
        'contrast': 'f_gain',
        'gamma': 'f_gamma',
        'conv_8_max': 'f_gain16_8',
        'hue': 'f_hue',
        }

#from VR doc: to be ignored, not used anymore
TO_BE_IGNORED_FIELDS = {
        'contrast_16': 'res_7',
        'bright_16': 'res_8',
        'rotate_16': 'res_9',
        'time_annotation': 'res_10',
        'trig_cine': 'res_11',
        'shutter_on': 'res_12',
        'binning': 'res_13',
        'b_mains_freq': 'res_14', 
        'b_time_code': 'res_15',
        'b_priority': 'res_16',
        'w_leap_sec_dy': 'res_17',
        'd_delay_tc': 'res_18',
        'd_delay_pps': 'res_19',
        'gen_bits': 'res_20',
        'conv_8_min': '',
        }

# from VR doc: last setup field appearing in software version
# TODO: keep up-to-date with newer and more precise doc, if available
END_OF_SETUP = {
        551: 'software_version',
        552: 'recording_time_zone',
        578: 'rotate',
        605: 'b_stamp_time',
        606: 'mc_percent63',
        607: 'head_serial3',
        614: 'decimation',
        624: 'master_serial',
        625: 'sensor',
        631: 'frame_delay_ns',
        637: 'description',
        671: 'hue',
        691: 'lens_focal_length',
        693: 'f_gain16_8',
        701: 'f_tc_rate',
        702: 'cine_name',
        }


class Cine(FramesSequence):
    """Read cine files

    Read cine files, the out put from Vision Research high-speed phantom
    cameras.  Support uncompressed monochrome and color files.

    Nominally thread-safe, but this assertion is not tested.


    Parameters
    ----------
    filename : string
        Path to cine (or chd) file.
    
    Notes
    -----
    For a .chd file, this class only reads the header, not the images.
    
    """
    # TODO: Unit tests using a small sample cine file.
    @classmethod
    def class_exts(cls):
        return {'cine'} | super(Cine, cls).class_exts()

    propagate_attrs = ['frame_shape', 'pixel_type', 'filename', 'frame_rate',
                       'get_fps', 'compression', 'cfa', 'off_set']

    def __init__(self, filename):
        super(Cine, self).__init__()
        self.f = open(filename, 'rb')
        self._filename = filename

        ### HEADER
        self.header_dict = self._read_header(HEADER_FIELDS)
        self.bitmapinfo_dict = self._read_header(BITMAP_INFO_FIELDS,
                                                self.off_image_header)
        self.setup_fields_dict = self._read_header(SETUP_FIELDS, self.off_setup)
        self.setup_fields_dict = self.clean_setup_dict()

        self._width = self.bitmapinfo_dict['bi_width']
        self._height = self.bitmapinfo_dict['bi_height']
        self._pixel_count = self._width * self._height

        # Allows Cine object to be accessed from multiple threads!
        self.file_lock = Lock()

        self._hash = None

        self._im_sz = (self._width, self._height)

        # sort out the data type by reading the meta-data
        if self.bitmapinfo_dict['bi_bit_count'] in (8, 24):
            self._data_type = 'u1'
        else:
            self._data_type = 'u2'
        self.tagged_blocks = self._read_tagged_blocks()
        self.frame_time_stamps = self.tagged_blocks['image_time_only']
        self.all_exposures = self.tagged_blocks['exposure_only']
        self.stack_meta_data = dict()
        self.stack_meta_data.update(self.bitmapinfo_dict)
        self.stack_meta_data.update({k: self.setup_fields_dict[k]
                                     for k in set(('trig_frame',
                                                   'gamma',
                                                   'frame_rate',
                                                   'shutter_ns'
                                                   )
                                                   )
                                                   })
        self.stack_meta_data.update({k: self.header_dict[k]
                                     for k in set(('first_image_no',
                                                   'image_count',
                                                   'total_image_count',
                                                   'first_movie_image'
                                                   )
                                                   )
                                                   })
        self.stack_meta_data['trigger_time'] = self.trigger_time

        ### IMAGES
        # Move to images offset to test EOF...
        self.f.seek(self.off_image_offsets)
        if self.f.read(1) != b'':
            # ... If no, read images
            self.image_locations = self._unpack('%dQ' % self.image_count,
                                               self.off_image_offsets)
            if type(self.image_locations) not in (list, tuple):
                self.image_locations = [self.image_locations]
        # TODO: add support for reading sequence within the same framework, when data
        # has been saved in another format (.tif, image sequence, etc)

    def clean_setup_dict(self):
        r"""Clean setup dictionary by removing newer fields, when compared to the
        software version, and trailing null character b'\x00' in entries.

        Notes
        -----
        The method is called after building the setup from the raw cine header.
        It can be overridden to match more specific purposes (e.g. filtering
        out TO_BE_IGNORED_ and UPDATED_FIELDS).

        See also
        --------
        `Vision Research Phantom documentation <http://phantomhighspeed-knowledge.force.com/servlet/fileField?id=0BE1N000000kD2i>`_
        """
        setup = self.setup_fields_dict.copy()
        # End setup at correct field (according to doc)
        versions = sorted(END_OF_SETUP.keys())
        fields = [v[0] for v in SETUP_FIELDS]
        v = setup['software_version']
        # Get next field where setup is known to have ended, according to VR
        try:
            v_up = versions[sorted(where(array(versions) >= v)[0])[0]]
            last_field = END_OF_SETUP[v_up]
            for k in fields[fields.index(last_field)+1:]:
                del setup[k]
        except IndexError:
            # Or go to the end (waiting for updated documentation)
            pass

        # Remove blank characters
        setup = _convert_null_byte(setup)

        # Filter out 'res_' (reserved/obsolete) fields
        #k_res = [k for k in setup.keys() if k.startswith('res_')]
        #for k in k_res:
        #    del setup[k]

        # Format f_tone properly
        if 'f_tone' in setup.keys():
            tone = setup['f_tone']
            setup['f_tone'] = tuple((tone[2*k], tone[2*k+1])\
                                    for k in range(setup['tone_points']))
        return setup


    @property
    def filename(self):
        return self._filename
    
    @property
    def frame_rate(self):
        """Frame rate (setting in Phantom PCC software) (Hz).
        May differ from computed average one.
        """
        return self.setup_fields_dict['frame_rate']

    @property
    def frame_rate_avg(self):
        """Actual frame rate, averaged on frame timestamps (Hz)."""
        return self.get_frame_rate_avg()

    # use properties for things that should not be changeable
    @property
    def cfa(self):
        return self.setup_fields_dict['cfa']

    @property
    def compression(self):
        return self.header_dict['compression']

    @property
    def pixel_type(self):
        return np.dtype(self._data_type)

    # TODO: what is this field??? (baneel)
    @property
    def off_set(self):
        return self.header_dict['offset']

    @property
    def setup_length(self):
        return self.setup_fields_dict['length']

    @property
    def off_image_offsets(self):
        return self.header_dict['off_image_offsets']

    @property
    def off_image_header(self):
        return self.header_dict['off_image_header']

    @property
    def off_setup(self):
        return self.header_dict['off_setup']

    @property
    def image_count(self):
        return self.header_dict['image_count']

    @property
    def frame_shape(self):
        return self._im_sz

    @property
    def shape(self):
        """Shape of virtual np.array containing images."""
        W, H = self.frame_shape
        return self.len(), H, W

    def get_frame(self, j):
        md = dict()
        md['exposure'] = self.all_exposures[j]
        ts, sec_frac = self.frame_time_stamps[j]
        md['frame_time'] = {'datetime': ts,
                            'second_fraction': sec_frac,
                            'time_to_trigger': self.get_time_to_trigger(j),
                            }
        return Frame(self._get_frame(j), frame_no=j, metadata=md)

    def _unpack(self, fs, offset=None):
        if offset is not None:
            self.f.seek(offset)
        s = _build_struct(fs)
        vals = s.unpack(self.f.read(s.size))
        if len(vals) == 1:
            return vals[0]
        else:
            return vals

    def _read_tagged_blocks(self):
        """Reads the tagged block meta-data from the header."""
        tmp_dict = dict()
        if not self.off_setup + self.setup_length < self.off_image_offsets:
            return
        next_tag_exists = True
        next_tag_offset = 0
        while next_tag_exists:
            block_size, next_tag_exists = self._read_tag_block(next_tag_offset,
                                                               tmp_dict)
            next_tag_offset += block_size
        return tmp_dict

    def _read_tag_block(self, off_set, accum_dict):
        '''
        Internal helper-function for reading the tagged blocks.
        '''
        with FileLocker(self.file_lock):
            self.f.seek(self.off_setup + self.setup_length + off_set)
            block_size = self._unpack(UINT32)
            b_type = self._unpack(UINT16)
            more_tags = self._unpack(UINT16)

            if b_type == 1004:
                # docs say to ignore range data it seems to be a poison flag,
                # if see this, give up tag parsing
                return block_size, 0

            try:
                d_name, d_type = TAGGED_FIELDS[b_type]

            except KeyError:
                return block_size, more_tags

            if d_type == '':
                # print "can't deal with  <" + d_name + "> tagged data"
                return block_size, more_tags

            s_tmp = _build_struct(d_type)
            if (block_size-8) % s_tmp.size != 0:
                #            print 'something is wrong with your data types'
                return block_size, more_tags

            d_count = (block_size-8)//(s_tmp.size)

            data = self._unpack('%d' % d_count + d_type)
            if not isinstance(data, tuple):
                # fix up data due to design choice in self.unpack
                data = (data, )

            # parse time
            if b_type == 1002 or b_type == 1001:
                data = [(datetime.datetime.fromtimestamp(d >> 32),
                         (FRACTION_MASK & d)/MAX_INT) for d in data]
            # convert exposure to seconds
            if b_type == 1003:
                data = [d/(MAX_INT) for d in data]

            accum_dict[d_name] = data

        return block_size, more_tags

    def _read_header(self, fields, offset=0):
        self.f.seek(offset)
        tmp = dict()
        for name, format in fields:
            val = self._unpack(format)
            tmp[name] = val

        return tmp

    def _get_frame(self, number):
        with FileLocker(self.file_lock):
            # get basic information about the frame we want
            image_start = self.image_locations[number]
            annotation_size = self._unpack(UINT32, image_start)
            # this is not used, but is needed to advance the point in the file
            annotation = self._unpack('%db' % (annotation_size - 8))
            image_size = self._unpack(UINT32)

            cfa = self.cfa
            compression = self.compression

            # sort out data type looking at the cached version
            data_type = self._data_type

            # actual bit per pixel
            actual_bits = image_size * 8 // (self._pixel_count)

            # so this seem wrong as 10 or 12 bits won't fit in 'u1'
            # but I (TAC) may not understand and don't have a packed file
            # (which the docs seem to imply don't exist) to test on so
            # I am leaving it.  good luck.
            if actual_bits in (10, 12):
                data_type = 'u1'

            # move the file to the right point in the file
            self.f.seek(image_start + annotation_size)

            # suck the data out of the file and shove into linear
            # numpy array
            frame = frombuffer(self.f.read(image_size), data_type)

            # if mono-camera
            if cfa == CFA_NONE:
                if compression != 0:
                    raise ValueError("Can not deal with compressed files\n" +
                                     "compression level: " +
                                     "{}".format(compression))
                # we are working with a monochrome camera
                # un-pack packed data
                if (actual_bits == 10):
                    frame = _ten2sixteen(frame)
                elif (actual_bits == 12):
                    frame = _twelve2sixteen(frame)
                elif (actual_bits % 8):
                    raise ValueError('Data should be byte aligned, ' +
                         'or 10 or 12 bit packed (appears to be' +
                        ' %dbits/pixel?!)' % actual_bits)

                # re-shape to an array
                # flip the rows
                frame = frame.reshape(self._height, self._width)[::-1]

                if actual_bits in (10, 12):
                    frame = frame[::-1, :]
                    # Don't know why it works this way, but it does...
            # else, some sort of color layout
            else:
                if compression == 0:
                    # and re-order so color is RGB (naively saves as BGR)
                    frame = frame.reshape(self._height, self._width,
                                          3)[::-1, :, ::-1]
                elif compression == 2:
                    raise ValueError("Can not process un-interpolated movies")
                else:
                    raise ValueError("Should never hit this, " +
                                     "you have an un-documented file\n" +
                                     "compression level: " +
                                     "{}".format(compression))

        return frame

    def __len__(self):
        return self.image_count

    len = __len__

    @index_attr
    def get_time(self, i):
        """Return the time of frame i in seconds, relative to first frame."""
        warnings.warn("This is not guaranteed to be the actual time. "\
                      +"See self.get_time_to_trigger(i) method.",
                      category=PendingDeprecationWarning)
        return float(i) / self.frame_rate

    @index_attr
    def get_time_to_trigger(self, i):
        """Get actual time (s) of frame i, relative to trigger."""
        ti = self.frame_time_stamps[i]
        ti = ti[0].timestamp() + ti[1]
        tt= self.trigger_time
        tt = tt['datetime'].timestamp() + tt['second_fraction']
        return ti - tt

    def get_frame_rate_avg(self, error_tol=1e-3):
        """Compute mean frame rate (Hz), on the basis of frame time stamps.

        Parameters
        ----------
        error_tol : float, optional.
            Tolerance on relative error (standard deviation/mean),
            above which a warning is raised.

        Returns
        -------
        fps : float.
            Actual mean frame rate, based on the frames time stamps.
        """
        times = np.r_[[self.get_time_to_trigger(i) for i in range(self.len())]]
        freqs = 1 / np.diff(times)
        fps, std = freqs.mean(), freqs.std()
        error = std / fps
        if error > error_tol:
            warnings.warn('Relative precision on the average frame rate is '\
                          +'{:.2f}%.'.format(1e2*error))
        return fps

    def get_fps(self):
        """Get frame rate (setting in Phantom PCC software) (Hz).
        May differ from computed average one.

        See also
        --------
        PCC setting (all fields refer to the same value)
            self.frame_rate
            self.setup_fields_dict['frame_rate']
        
        Computed average
            self.frame_rate_avg
            self.get_frame_rate_avg()
        """
        return self.frame_rate

    def close(self):
        self.f.close()

    def __unicode__(self):
        return self.filename

    def __str__(self):
        return unicode(self).encode('utf-8')

    def __repr__(self):
        # May be overwritten by subclasses
        return """<Frames>
Source: {filename}
Length: {count} frames
Frame Shape: {frame_shape!r}
Pixel Datatype: {dtype}""".format(frame_shape=self.frame_shape,
                                  count=len(self),
                                  filename=self.filename,
                                  dtype=self.pixel_type)

    @property
    def trigger_time(self):
        '''Returns the time of the trigger, tuple of (datatime_object,
        fraction_in_s)'''
        trigger_time = self.header_dict['trigger_time']
        ts, sf = (datetime.datetime.fromtimestamp(trigger_time >> 32),
                   float(FRACTION_MASK & trigger_time)/(MAX_INT))

        return {'datetime': ts, 'second_fraction': sf}

    @property
    def hash(self):
        if self._hash is None:
            self._hash_fun()
        return self._hash

    def __hash__(self):
        return int(self.hash, base=16)

    def _hash_fun(self):
        """Generates the md5 hash of the header of the file.  Here the
        header is defined as everything before the first image starts.

        This includes all of the meta-data (including the plethora of
        time stamps) so this will be unique.
        """
        # get the file lock (so we don't screw up any other reads)
        with FileLocker(self.file_lock):

            self.f.seek(0)
            max_loc = self.image_locations[0]
            md5 = hashlib.md5()

            chunk_size = 128*md5.block_size
            chunk_count = (max_loc//chunk_size) + 1

            for j in range(chunk_count):
                md5.update(self.f.read(128*md5.block_size))

            self._hash = md5.hexdigest()

    def __eq__(self, other):
        return self.hash == other.hash

    def __ne__(self, other):
        return not self == other


# Should be divisible by 3, 4 and 5!  This seems to be near-optimal.
CHUNK_SIZE = 6 * 10 ** 5


def _ten2sixteen(a):
    """Convert array of 10bit uints to array of 16bit uints."""
    b = np.zeros(a.size//5*4, dtype='u2')

    for j in range(0, len(a), CHUNK_SIZE):
        (a0, a1, a2, a3, a4) = [a[j+i:j+CHUNK_SIZE:5].astype('u2')
                                for i in range(5)]

        k = j//5 * 4
        k2 = k + CHUNK_SIZE//5 * 4

        b[k+0:k2:4] = ((a0 & 0b11111111) << 2) + ((a1 & 0b11000000) >> 6)
        b[k+1:k2:4] = ((a1 & 0b00111111) << 4) + ((a2 & 0b11110000) >> 4)
        b[k+2:k2:4] = ((a2 & 0b00001111) << 6) + ((a3 & 0b11111100) >> 2)
        b[k+3:k2:4] = ((a3 & 0b00000011) << 8) + ((a4 & 0b11111111) >> 0)

    return b


def _sixteen2ten(b):
    """Convert array of 16bit uints to array of 10bit uints."""
    a = np.zeros(b.size//4*5, dtype='u1')

    for j in range(0, len(a), CHUNK_SIZE):
        (b0, b1, b2, b3) = [b[j+i:j+CHUNK_SIZE:4] for i in range(4)]

        k = j//4 * 5
        k2 = k + CHUNK_SIZE//4 * 5

        a[k+0:k2:5] =                              ((b0 & 0b1111111100) >> 2)
        a[k+1:k2:5] = ((b0 & 0b0000000011) << 6) + ((b1 & 0b1111110000) >> 4)
        a[k+2:k2:5] = ((b1 & 0b0000001111) << 4) + ((b2 & 0b1111000000) >> 6)
        a[k+3:k2:5] = ((b2 & 0b0000111111) << 2) + ((b3 & 0b1100000000) >> 8)
        a[k+4:k2:5] = ((b3 & 0b0011111111) << 0)

    return a


def _twelve2sixteen(a):
    """Convert array of 12bit uints to array of 16bit uints."""
    b = np.zeros(a.size//3*2, dtype='u2')

    for j in range(0, len(a), CHUNK_SIZE):
        (a0, a1, a2) = [a[j+i:j+CHUNK_SIZE:3].astype('u2') for i in range(3)]

        k = j//3 * 2
        k2 = k + CHUNK_SIZE//3 * 2

        b[k+0:k2:2] = ((a0 & 0xFF) << 4) + ((a1 & 0xF0) >> 4)
        b[k+1:k2:2] = ((a1 & 0x0F) << 8) + ((a2 & 0xFF) >> 0)

    return b


def _sixteen2twelve(b):
    """Convert array of 16bit uints to array of 12bit uints."""
    a = np.zeros(b.size//2*3, dtype='u1')

    for j in range(0, len(a), CHUNK_SIZE):
        (b0, b1) = [b[j+i:j+CHUNK_SIZE:2] for i in range(2)]

        k = j//2 * 3
        k2 = k + CHUNK_SIZE//2 * 3

        a[k+0:k2:3] =                       ((b0 & 0xFF0) >> 4)
        a[k+1:k2:3] = ((b0 & 0x00F) << 4) + ((b1 & 0xF00) >> 8)
        a[k+2:k2:3] = ((b1 & 0x0FF) << 0)

    return a


def _convert_null_byte(dic):
    """
    Convert binary null character b'\x00' to empty string in dictionary entries.

    Parameters
    ----------
    dic : dict
        Dictionary to clean. Function loops over the key-value pairs and
        converts the null byte `b'\x00'` to empty string `''`.

    Returns
    -------
    clean_dic : dict
        Cleaned dictionary.

    Notes
    -----
    The routine is intended to work on string-like bytes array (resp. Iterable
    of such arrays), and return a string (resp. a list of strings).
    """
    for k, v in dic.items():
        if isinstance(v, bytes):
            try:
                dic[k] = v.decode('utf8').replace('\x00', '')
            except (UnicodeDecodeError):
                pass
        elif isinstance(v, Iterable):
            try:
                dic[k] = [el.decode('utf8').replace('\x00', '')\
                          for el in v]
            except (AttributeError, UnicodeDecodeError):
                pass
    return dic



import os
import tempfile
import zipfile
import unittest
import numpy as np
from numpy.testing import (assert_equal, assert_allclose)
import pims

from pims.tests.test_common import (_image_series,
                                    clean_dummy_png, save_dummy_png,
                                    _skip_if_no_skimage)

path, _ = os.path.split(os.path.abspath(__file__))
path = os.path.join(path, 'data')




class TestImageSequenceWithPIL(_image_series, unittest.TestCase):
    def setUp(self):
        _skip_if_no_skimage()
        self.filepath = os.path.join(path, 'image_sequence')
        self.filenames = ['T76S3F00001.png', 'T76S3F00002.png',
                          'T76S3F00003.png', 'T76S3F00004.png',
                          'T76S3F00005.png']
        shape = (10, 11)
        frames = save_dummy_png(self.filepath, self.filenames, shape)

        self.filename = os.path.join(self.filepath, '*.png')
        self.frame0 = frames[0]
        self.frame1 = frames[1]
        self.kwargs = dict(plugin='pil')
        self.klass = pims.ImageSequence
        self.v = self.klass(self.filename, **self.kwargs)
        self.expected_shape = shape
        self.expected_len = 5
        self.tempdir = tempfile.mkdtemp()
        self.tempfile = os.path.join(self.tempdir, 'test.zip')

        with zipfile.ZipFile(self.tempfile, 'w') as archive:
            for fn in self.filenames:
                archive.write(os.path.join(self.filepath, fn))

    def test_bad_path_raises(self):
        raises = lambda: pims.ImageSequence('this/path/does/not/exist/*.jpg')
        self.assertRaises(IOError, raises)

    def test_zipfile(self):
        pims.ImageSequence(self.tempfile)[0]

    def tearDown(self):
        clean_dummy_png(self.filepath, self.filenames)
        os.remove(self.tempfile)
        os.rmdir(self.tempdir)


class TestImageSequenceWithMPL(_image_series, unittest.TestCase):
    def setUp(self):
        _skip_if_no_skimage()
        self.filepath = os.path.join(path, 'image_sequence')
        self.filenames = ['T76S3F00001.png', 'T76S3F00002.png',
                          'T76S3F00003.png', 'T76S3F00004.png',
                          'T76S3F00005.png']
        shape = (10, 11)
        frames = save_dummy_png(self.filepath, self.filenames, shape)
        self.filename = os.path.join(self.filepath, '*.png')
        self.frame0 = frames[0]
        self.frame1 = frames[1]
        self.kwargs = dict(plugin='matplotlib')
        self.klass = pims.ImageSequence
        self.v = self.klass(self.filename, **self.kwargs)
        self.expected_shape = shape
        self.expected_len = 5

    def tearDown(self):
        clean_dummy_png(self.filepath, self.filenames)


class TestImageSequenceAcceptsList(_image_series, unittest.TestCase):
    def setUp(self):
        self.filepath = os.path.join(path, 'image_sequence')
        self.filenames = ['T76S3F00001.png', 'T76S3F00002.png',
                          'T76S3F00003.png', 'T76S3F00004.png',
                          'T76S3F00005.png']
        shape = (10, 11)
        frames = save_dummy_png(self.filepath, self.filenames, shape)

        self.filename = [os.path.join(self.filepath, fn)
                         for fn in self.filenames]
        self.frame0 = frames[0]
        self.frame1 = frames[1]
        self.kwargs = dict(plugin='matplotlib')
        self.klass = pims.ImageSequence
        self.v = self.klass(self.filename, **self.kwargs)
        self.expected_shape = shape
        self.expected_len = len(self.filenames)

    def tearDown(self):
        clean_dummy_png(self.filepath, self.filenames)


class TestImageSequenceNaturalSorting(_image_series, unittest.TestCase):
    def setUp(self):
        self.filepath = os.path.join(path, 'image_sequence')
        self.filenames = ['T76S3F1.png', 'T76S3F20.png',
                          'T76S3F3.png', 'T76S3F4.png',
                          'T76S3F50.png', 'T76S3F10.png']
        shape = (10, 11)
        frames = save_dummy_png(self.filepath, self.filenames, shape)

        self.filename = os.path.join(self.filepath, 'T76*.png')
        self.frame0 = frames[0]
        self.frame1 = frames[2]
        self.kwargs = dict(plugin='matplotlib')
        self.klass = pims.ImageSequence
        self.v = self.klass(self.filename, **self.kwargs)
        self.expected_shape = shape
        self.expected_len = len(self.filenames)

    def test_sorting(self):
        sorted_files = ['T76S3F1.png',
                        'T76S3F3.png',
                        'T76S3F4.png',
                        'T76S3F10.png',
                        'T76S3F20.png',
                        'T76S3F50.png']

        assert sorted_files == [x.split(os.path.sep)[-1]
                                for x in self.v._filepaths]

        # provide a list: there should be no sorting
        self.filename = [os.path.join(self.filepath, fn) for fn in
                         self.filenames]
        v_unsorted = self.klass(self.filename, **self.kwargs)

        assert self.filenames == [x.split(os.path.sep)[-1]
                                  for x in v_unsorted._filepaths]


    def tearDown(self):
        clean_dummy_png(self.filepath, self.filenames)


class ImageSequenceND(_image_series, unittest.TestCase):
    def setUp(self):
        self.filepath = os.path.join(path, 'image_sequence3d')
        self.filenames = ['file_t001_z001_c1.png',
                          'file_t001_z001_c2.png',
                          'file_t001_z002_c1.png',
                          'file_t001_z002_c2.png',
                          'file_t002_z001_c1.png',
                          'file_t002_z001_c2.png',
                          'file_t002_z002_c1.png',
                          'file_t002_z002_c2.png',
                          'file_t003_z001_c1.png',
                          'file_t003_z001_c2.png',
                          'file_t003_z002_c1.png',
                          'file_t003_z002_c2.png']
        shape = (10, 11)
        frames = save_dummy_png(self.filepath, self.filenames, shape)

        self.filename = os.path.join(self.filepath, '*.png')
        self.frame0 = np.array([frames[0], frames[2]])
        self.frame1 = np.array([frames[4], frames[6]])
        self.klass = pims.ImageSequenceND
        self.kwargs = dict(axes_identifiers='tzc')
        self.v = self.klass(self.filename, **self.kwargs)
        self.v.default_coords['c'] = 0
        self.expected_len = 3
        self.expected_Z = 2
        self.expected_C = 2
        self.expected_shape = (self.expected_Z,) + shape

    def tearDown(self):
        clean_dummy_png(self.filepath, self.filenames)

    def test_filename_tzc(self):
        tzc = pims.image_sequence.filename_to_indices('file_t01_z005_c4.png')
        self.assertEqual(tzc, [1, 5, 4])
        tzc = pims.image_sequence.filename_to_indices('t01file_t01_z005_c4.png')
        self.assertEqual(tzc, [1, 5, 4])
        tzc = pims.image_sequence.filename_to_indices('file_z005_c4_t01.png')
        self.assertEqual(tzc, [1, 5, 4])
        tzc = pims.image_sequence.filename_to_indices(u'file\u03BC_z05_c4_t01.png')
        self.assertEqual(tzc, [1, 5, 4])
        tzc = pims.image_sequence.filename_to_indices('file_t9415_z005.png')
        self.assertEqual(tzc, [9415, 5, 0])
        tzc = pims.image_sequence.filename_to_indices('file_t47_c34.png')
        self.assertEqual(tzc, [47, 0, 34])
        tzc = pims.image_sequence.filename_to_indices('file_z4_c2.png')
        self.assertEqual(tzc, [0, 4, 2])
        tzc = pims.image_sequence.filename_to_indices('file_p4_c2_q5_r1.png',
                                                      ['p', 'q', 'r'])
        self.assertEqual(tzc, [4, 5, 1])

    def test_sizeZ(self):
        self.check_skip()
        assert_equal(self.v.sizes['z'], self.expected_Z)

    def test_sizeC(self):
        self.check_skip()
        assert_equal(self.v.sizes['c'], self.expected_C)


class ImageSequenceND_RGB(_image_series, unittest.TestCase):
    def setUp(self):
        self.filepath = os.path.join(path, 'image_sequence3d')
        self.filenames = ['file_t001_z001_c1.png',
                          'file_t001_z002_c1.png',
                          'file_t002_z001_c1.png',
                          'file_t002_z002_c1.png',
                          'file_t003_z001_c1.png',
                          'file_t003_z002_c1.png']
        shape = (10, 11, 3)
        frames = save_dummy_png(self.filepath, self.filenames, shape)

        self.filename = os.path.join(self.filepath, '*.png')
        self.frame0 = np.array([frames[0][:, :, 0], frames[1][:, :, 0]])
        self.frame1 = np.array([frames[2][:, :, 0], frames[3][:, :, 0]])
        self.klass = pims.ImageSequenceND
        self.kwargs = dict(axes_identifiers='tz')
        self.v = self.klass(self.filename, **self.kwargs)
        self.v.default_coords['c'] = 0
        self.expected_len = 3
        self.expected_Z = 2
        self.expected_C = 3
        self.expected_shape = (2, 10, 11)

    def test_sizeZ(self):
        self.check_skip()
        assert_equal(self.v.sizes['z'], self.expected_Z)

    def test_sizeC(self):
        self.check_skip()
        assert_equal(self.v.sizes['c'], self.expected_C)

    def tearDown(self):
        clean_dummy_png(self.filepath, self.filenames)


class ReaderSequence(_image_series, unittest.TestCase):
    def setUp(self):
        self.filepath = os.path.join(path, 'image_sequence3d')
        self.filenames = ['file001.png',
                          'file002.png',
                          'file003.png',
                          'file004.png']
        shape = (10, 11, 3)
        frames = save_dummy_png(self.filepath, self.filenames, shape)

        self.filename = os.path.join(self.filepath, '*.png')
        self.frame0 = frames[0]
        self.frame1 = frames[1]
        self.klass = pims.ReaderSequence
        self.kwargs = dict(reader_cls=pims.ImageReaderND, axis_name='t')
        self.v = self.klass(self.filename, **self.kwargs)
        self.v.bundle_axes = 'yxc'
        self.v.iter_axes = 't'
        self.expected_len = 4
        self.expected_C = 3
        self.expected_shape = (10, 11, 3)

    def test_sizeC(self):
        self.check_skip()
        assert_equal(self.v.sizes['c'], self.expected_C)

    def tearDown(self):
        clean_dummy_png(self.filepath, self.filenames)



import re

__all__ = ["natural_keys"]


def _atoi(text):
    return int(text) if text.isdigit() else text


def natural_keys(text):
    """Sort list of string in a human way.
    See: http://stackoverflow.com/questions/5967500/how-to-correctly-sort-a-string-with-a-number-
    inside
    Examples
    --------
    >>> alist=["something1", "something12", "something17", "something2"]
    >>> alist.sort(key=natural_keys)
    >>> print(alist)
    ['something1', 'something2', 'something12', 'something17']
    """
    return [_atoi(c) for c in re.split(r'(\d+)', text)]



import numpy as np

from pims.base_frames import FramesSequence, FramesSequenceND
from pims.frame import Frame

try:
    from skimage.io import imread
except ImportError:
    import imageio

    def imread(*args, **kwargs):  # Strip metadata for consistency.
        return np.asarray(imageio.imread(*args, **kwargs))


class ImageReader(FramesSequence):
    """Reads a single image into a length-1 reader.

    Simple wrapper around skimage.io.imread or matplotlib.pyplot.imread,
    in that priority order."""
    @classmethod
    def class_exts(cls):
        return {'png', 'jpg', 'jpeg', 'gif', 'bmp', 'ico'}

    class_priority = 12

    def __init__(self, filename, **kwargs):
        self._data = imread(filename, **kwargs)

    def get_frame(self, i):
        return Frame(self._data, frame_no=0)

    def __len__(self):
        return 1

    @property
    def pixel_type(self):
        return self._data.dtype

    @property
    def frame_shape(self):
        return self._data.shape


class ImageReaderND(FramesSequenceND):
    """Reads a single image into a dimension-aware reader.

    Simple wrapper around skimage.io.imread or matplotlib.pyplot.imread,
    in that priority order."""
    @classmethod
    def class_exts(cls):
        return {'png', 'jpg', 'jpeg', 'gif', 'bmp', 'ico'}

    class_priority = 11

    def __init__(self, filename, **kwargs):
        super(ImageReaderND, self).__init__()

        self._data = Frame(imread(filename, **kwargs), frame_no=0)
        shape = self._data.shape
        if len(shape) == 2:   # greyscale
            self._init_axis('y', shape[0])
            self._init_axis('x', shape[1])
            self._register_get_frame(self.get_frame_2D, 'yx')
            self.bundle_axes = 'yx'
        elif (len(shape) == 3) and (shape[2] in (3, 4)):   # grayscale
            self._init_axis('y', shape[0])
            self._init_axis('x', shape[1])
            self._init_axis('c', shape[2])
            self._register_get_frame(self.get_frame_2D, 'yxc')
            self.bundle_axes = 'yxc'
        else:
            raise IOError('The image has a shape that is not grayscale nor RGB:'
                          ' {}'.format(shape))

    def get_frame_2D(self, **ind):
        return Frame(self._data, frame_no=0)

    @property
    def pixel_type(self):
        return self._data.dtype

    @property
    def frame_shape(self):
        return self._data.shape



# Tests for cine.py

import os
from datetime import datetime
import unittest
import numpy as np
import pims
import pims.cine

tests_path, _ = os.path.split(os.path.abspath(__file__))

class _common_cine_sample_tests(object):
    """Test the .cine reader on a sample file."""
    def setUp(self):
        self.cin = pims.open(self.sample_filename, **self.options)

    def tearDown(self):
        self.cin.close()

    def test_type(self):
        assert isinstance(self.cin, pims.cine.Cine)

    def test_metadata_smoke(self):
        assert len(self.cin)
        assert self.cin.frame_rate
        assert len(self.cin.frame_shape) == 2


class test_legacy_cine_sample(_common_cine_sample_tests, unittest.TestCase):
    # File made by Nathan Keim in December 2008, using Phantom Camera Control
    def setUp(self):
        self.sample_filename = os.path.join(tests_path, 'data',
                                            'cine_legacy.cine')
        if not os.path.exists(self.sample_filename):
            raise unittest.SkipTest('Legacy sample cine file not found. '
                                    'Skipping.')

        self.options = {}
        super(test_legacy_cine_sample, self).setUp()

    def test_length(self):
        assert len(self.cin) == 97


class test_new_cine_sample(_common_cine_sample_tests, unittest.TestCase):
    # File made by Baptiste Neel in Feb. 2019, using Phantom Camera Control v3.3.781.0
    # Recording fps 90,000, with decimation 10 at saving (actual fps 9,000)
    """Tests for the newer format only"""
    def setUp(self):
        self.sample_filename = os.path.join(tests_path, 'data',
                                            'cine_781.cine')
        if not os.path.exists(self.sample_filename):
            raise unittest.SkipTest('Newer sample cine file not found. '
                                    'Skipping.')
        self.options = {}
        super(test_new_cine_sample, self).setUp()

    def test_new_file_info(self):
        """Tests based on the specific file in the repo."""
        c = self.cin
        pass



import unittest
import os
import functools
import numpy as np
import pims
from numpy.testing import assert_array_equal
from pims import plot_to_frame, plots_to_frame
from pims.display import export_moviepy, export_pyav
from .test_common import _skip_if_no_MoviePy, _skip_if_no_PyAV, path

import unittest

try:
    import matplotlib as mpl
    import matplotlib.pyplot as plt
except ImportError:
    mpl = None
    plt = None


def _skip_if_no_mpl():
    if plt is None:
        raise unittest.SkipTest('Matplotlib not installed. Skipping.')


class TestPlotToFrame(unittest.TestCase):
    def setUp(self):
        _skip_if_no_mpl()
        plt.switch_backend('Agg')  # does not plot to screen
        x = np.linspace(0, 2 * np.pi, 100)
        t = np.linspace(0, 2 * np.pi, 10)
        y = np.sin(x[np.newaxis, :] - t[:, np.newaxis])

        self.figures = []
        self.axes = []
        for line in y:
            fig = plt.figure(figsize=(8, 6), tight_layout=False)
            ax = fig.gca()
            ax.plot(x, line)
            self.figures.append(fig)
            self.axes.append(ax)

        self.expected_shape = (10, 384, 512, 4)

        mpl.rcParams.update({'savefig.bbox': None})

    def tearDown(self):
        for fig in self.figures:
            plt.close(fig)

    def test_ax_to_frame(self):
        frame = plot_to_frame(self.axes[0])
        self.assertEqual(frame.shape, (384, 512, 4))

    def test_plot_to_frame(self):
        frame = plot_to_frame(self.figures[0])
        self.assertEqual(frame.shape, (384, 512, 4))

    def test_axes_to_frame(self):
        frame = plots_to_frame(self.axes)
        self.assertEqual(frame.shape, (10, 384, 512, 4))

    def test_plots_to_frame(self):
        frame = plots_to_frame(self.figures)
        self.assertEqual(frame.shape, (10, 384, 512, 4))

    def test_plot_width(self):
        width = np.random.randint(100, 1000)
        frame = plot_to_frame(self.figures[0], width)
        self.assertEqual(frame.shape[1], width)

    def test_plot_tight(self):
        fig = self.figures[0]
        fig.set_tight_layout(False)  # default to standard
        self.assertEqual(plot_to_frame(fig).shape[:2], (384, 512))
        self.assertLess(
            plot_to_frame(fig, bbox_inches='tight').shape[:2], (384, 512))
        self.assertEqual(plot_to_frame(fig).shape[:2], (384, 512))

        fig.set_tight_layout(True)  # default to tight
        self.assertLess(plot_to_frame(fig).shape[:2], (384, 512))
        self.assertEqual(
            plot_to_frame(fig, bbox_inches='standard').shape[:2], (384, 512))
        self.assertLess(plot_to_frame(fig).shape[:2], (384, 512))

    def test_plots_tight(self):
        frame = plots_to_frame(self.figures, bbox_inches='tight')
        self.assertLess(frame.shape[1:3], (384, 512))

    def test_plot_resize(self):
        frame = plot_to_frame(self.figures[0], fig_size_inches=(4, 4))
        self.assertEqual(frame.shape, (512, 512, 4))

    def test_plots_resize(self):
        frame = plots_to_frame(self.figures, fig_size_inches=(4, 4))
        self.assertEqual(frame.shape, (10, 512, 512, 4))

    def test_plots_width(self):
        width = np.random.randint(100, 1000)
        frame = plots_to_frame(self.figures, width)
        self.assertEqual(frame.shape[2], width)

    def test_plots_from_generator(self):
        frame = plots_to_frame(iter(self.figures))
        self.assertEqual(frame.shape, (10, 384, 512, 4))


class ExportCommon(object):
    def setUp(self):
        self.frame0 = np.load(os.path.join(path, 'bulk-water_frame0.npy'))
        self.frame1 = np.load(os.path.join(path, 'bulk-water_frame1.npy'))
        h, w = 128, 128
        self.expected_shape = (h, w, 3)
        self.expected_shape_rgba = (h, w, 4)
        self.expected_len = 20

        self.sequence = np.random.randint(0, 255,
                                          size=(self.expected_len,) +
                                               self.expected_shape,
                                          ).astype(np.uint8)
        self.sequence_rgba = np.random.randint(0, 255,
                                               size=(self.expected_len,) +
                                                    self.expected_shape_rgba,
                                               ).astype(np.uint8)
        self.tempfile = 'tempvideo.avi'  # avi containers support most codecs

    def tearDown(self):
        if hasattr(self, 'tempfile') and os.path.isfile(self.tempfile):
            os.remove(self.tempfile)

    def test_quality_wmv2(self):
        self.export_func(self.sequence, self.tempfile, codec='wmv2',
                         quality=1)
        lossless_size = int(os.path.getsize(self.tempfile))

        self.export_func(self.sequence, self.tempfile, codec='wmv2',
                         quality=0.01)
        compressed_size = int(os.path.getsize(self.tempfile))

        self.assertLess(compressed_size, lossless_size)

    def test_quality_mpeg4(self):
        self.export_func(self.sequence, self.tempfile, codec='mpeg4',
                         quality=1)
        lossless_size = int(os.path.getsize(self.tempfile))

        self.export_func(self.sequence, self.tempfile, codec='mpeg4',
                         quality=5)
        compressed_size = int(os.path.getsize(self.tempfile))

        self.assertLess(compressed_size, lossless_size)

    def test_quality_h264(self):
        self.export_func(self.sequence, self.tempfile, codec='libx264',
                         quality=0)
        lossless_size = int(os.path.getsize(self.tempfile))

        self.export_func(self.sequence, self.tempfile, codec='libx264',
                         quality=23)
        compressed_size = int(os.path.getsize(self.tempfile))

        self.assertLess(compressed_size, lossless_size)

    def test_rgba_h264(self):
        """Remove alpha channel."""
        # Start with smoke test for H.264
        self.export_func(self.sequence_rgba, self.tempfile, codec='libx264')
        # Check that RGB channels are preserved
        self.export_func(self.sequence_rgba, self.tempfile, codec='rawvideo',
                         pixel_format='bgr24')
        sequence_rgba_stripped = self.sequence_rgba[:, :, :, :3]
        with pims.open(self.tempfile) as reader:
            self.assertEqual(len(reader), self.expected_len)
            self.assertEqual(reader.frame_shape, self.expected_shape)
            for a, b in zip(sequence_rgba_stripped, reader):
                assert_array_equal(a, b)

    def test_rawvideo_export(self):
        """Exported frames must equal the input exactly"""
        self.export_func(self.sequence, self.tempfile, codec='rawvideo',
                         pixel_format='bgr24')
        with pims.open(self.tempfile) as reader:
            self.assertEqual(len(reader), self.expected_len)
            self.assertEqual(reader.frame_shape, self.expected_shape)
            for a, b in zip(self.sequence, reader):
                assert_array_equal(a, b)

    def test_rgb_memorder(self):
        """Fortran memory order must be converted."""
        self.export_func(self.sequence.astype(self.sequence.dtype,
                                              order='F'),
                         self.tempfile, codec='rawvideo',
                         pixel_format='bgr24')
        with pims.open(self.tempfile) as reader:
            self.assertEqual(len(reader), self.expected_len)
            self.assertEqual(reader.frame_shape, self.expected_shape)
            for a, b in zip(self.sequence, reader):
                assert_array_equal(a, b)


class TestExportMoviePy(ExportCommon, unittest.TestCase):
    def setUp(self):
        _skip_if_no_MoviePy()
        self.export_func = functools.partial(export_moviepy, verbose=False)
        ExportCommon.setUp(self)

    def tearDown(self):
        super(TestExportMoviePy, self).tearDown()


class TestExportPyAV(ExportCommon, unittest.TestCase):
    def setUp(self):
        _skip_if_no_PyAV()
        self.export_func = export_pyav
        ExportCommon.setUp(self)

    def tearDown(self):
        super(TestExportPyAV, self).tearDown()



from pims.base_frames import FramesSequenceND
from pims.frame import Frame

try:
    import imageio
except ImportError:
    imageio = None
try:
    import imageio_ffmpeg
    try:
        imageio_ffmpeg.get_ffmpeg_exe()
    except RuntimeError:
        imageio_ffmpeg = None
except ImportError:
    imageio_ffmpeg = None


def available():
    return imageio is not None


def ffmpeg_available():
    return (imageio is not None) and (imageio_ffmpeg is not None)


class ImageIOReader(FramesSequenceND):
    class_priority = 6

    propagate_attrs = ['frame_shape', 'pixel_type', 'metadata',
                       'get_metadata_raw', 'reader_class_name']

    @classmethod
    def class_exts(cls):
        exts = {'tiff', 'bmp', 'cut', 'dds', 'exr', 'g3', 'hdr', 'iff', 'j2k',
                'jng', 'jp2', 'jpeg', 'jpg', 'koala', 'pbm', 'pbmraw', 'pcd',
                'pcx', 'pfm', 'pgm', 'pgmraw', 'pict', 'png', 'ppm', 'ppmraw',
                'psd', 'ras', 'raw', 'sgi', 'targa', 'fi_tiff', 'wbmp', 'webp',
                'xbm', 'xpm', 'ico', 'gif', 'dicom', 'npz', 'fits', 'itk',
                'gdal', 'dummy', 'gif', 'ffmpeg', 'avbin', 'swf', 'fits',
                'gdal', 'ts', 'tif'}
        return exts.union(cls.additional_class_exts())

    @classmethod
    def additional_class_exts(cls):
        """If imageio-ffmpeg is available, more filetypes are supported."""
        movie_exts = set()
        if imageio_ffmpeg is not None:
            movie_exts = movie_exts.union(
                {'mov', 'avi', 'mpg', 'mpeg', 'mp4', 'mkv', 'wmv'}
            )
        return movie_exts

    def __init__(self, filename, **kwargs):
        if imageio is None:
            raise ImportError('The ImageIOReader requires imageio and '
                              '(for imageio >= 2.5) imageio-ffmpeg to work.')

        super(self.__class__, self).__init__()

        self.reader = imageio.get_reader(filename, **kwargs)
        self.filename = filename
        self._len = self.reader.get_length()
        # fallback to count_frames, for newer imageio versions
        if self._len == float("inf"):
            self._len = self.reader.count_frames()
        try:
            int(self._len)
        except OverflowError:
            self.reader.close()
            raise NotImplementedError(
                "Do not know how to deal with infinite readers"
                )

        first_frame = self.get_frame_2D(t=0)
        self._shape = first_frame.shape
        self._dtype = first_frame.dtype

        self._setup_axes()
        self._register_get_frame(self.get_frame_2D, 'yx')

    def _setup_axes(self):
        """Setup the xyctz axes, iterate over t axis by default

        """
        if self._shape[1] > 0:
            self._init_axis('x', self._shape[1])
        if self._shape[0] > 0:
            self._init_axis('y', self._shape[0])
        if self._len > 0:
            self._init_axis('t', self._len)

        if len(self.sizes) == 0:
            raise EmptyFileError("No axes were found for this file.")

        # provide the default
        self.iter_axes = self._guess_default_iter_axis()


    def _guess_default_iter_axis(self):
        """
        Guesses the default axis to iterate over based on axis sizes.
        Returns:
            the axis to iterate over
        """
        priority = ['t', 'z', 'c', 'v']
        found_axes = []
        for axis in priority:
            try:
                current_size = self.sizes[axis]
            except KeyError:
                continue

            if current_size > 1:
                return axis

            found_axes.append(axis)

        return found_axes[0]

    def get_frame_2D(self, **coords):
        i = coords['t'] if 't' in coords else 0
        frame = self.reader.get_data(i)
        return Frame(frame, frame_no=i, metadata=frame.meta)

    def get_metadata(self):
        return self.reader.get_meta_data(None)

    def __len__(self):
        return self._len

    def __iter__(self):
        iterable = self.reader.iter_data()
        for i in range(len(self)):
            frame = next(iterable)
            yield Frame(frame, frame_no=i, metadata=frame.meta)

    @property
    def frame_rate(self):
        return self.get_metadata()['fps']

    @property
    def frame_shape(self):
        return self._shape

    @property
    def pixel_type(self):
        return self._dtype

    def close(self):
        self.reader.close()
        super().close()



print("import: 'pims'")
import pims

print("import: 'pims.api'")
import pims.api

print("import: 'pims.base_frames'")
import pims.base_frames

print("import: 'pims.bioformats'")
import pims.bioformats

print("import: 'pims.cine'")
import pims.cine

print("import: 'pims.display'")
import pims.display

print("import: 'pims.ffmpeg_reader'")
import pims.ffmpeg_reader

print("import: 'pims.frame'")
import pims.frame

print("import: 'pims.image_reader'")
import pims.image_reader

print("import: 'pims.image_sequence'")
import pims.image_sequence

print("import: 'pims.imageio_reader'")
import pims.imageio_reader

print("import: 'pims.moviepy_reader'")
import pims.moviepy_reader

print("import: 'pims.norpix_reader'")
import pims.norpix_reader

print("import: 'pims.pyav_reader'")
import pims.pyav_reader

print("import: 'pims.spe_stack'")
import pims.spe_stack

print("import: 'pims.tiff_stack'")
import pims.tiff_stack

print("import: 'pims.tests'")
import pims.tests

print("import: 'pims.utils'")
import pims.utils

print("import: 'pims.utils.misc'")
import pims.utils.misc

print("import: 'pims.utils.sort'")
import pims.utils.sort




import numpy as np
import unittest
from numpy.testing import assert_array_equal
from pims.image_sequence import customize_image_sequence


def test_customize_image_sequence():
    dummy = lambda filename, **kwargs: np.zeros((1, 1))
    reader = customize_image_sequence(dummy)
    result = reader(['a', 'b', 'c'])
    actual_len = len(result)
    assert actual_len == 3
    for frame in result:
        assert_array_equal(frame, np.zeros((1, 1)))
    reader2 = customize_image_sequence(dummy, 'my_name')
    assert reader2.__name__ == 'my_name'



import numpy as np
import itertools
from slicerator import Slicerator, propagate_attr, index_attr
from .frame import Frame
from abc import ABC, abstractmethod, abstractproperty
from warnings import warn


class FramesStream(ABC):
    """
    A base class for wrapping input data which knows how to
    advance to the next frame, but does not have random access.

    The length does not need to be finite.

    Does not support slicing.
    """

    @abstractmethod
    def __iter__(self):
        pass

    @abstractproperty
    def pixel_type(self):
        """Returns a numpy.dtype for the data type of the pixel values"""
        pass

    @property
    def dtype(self):
        # The choice of using the separate name `pixel_type` was historical
        # and needlessly made PIMS objects look less like numpy arrays.
        return np.dtype(self.pixel_type)

    @abstractproperty
    def frame_shape(self):
        """Returns the shape of a single frame as a tuple ex (10, 12)"""
        pass

    @classmethod
    def class_exts(cls):
        """
        Return a set of the file extensions that this reader can deal with.

        Sub-classes should over-ride this function to list what extensions
        they deal with.

        The default interpretation of the returned set is 'file
        extensions including but not exclusively'.
        """
        return set()

    @property
    def exts(self):
        """
        Property to get the extensions of a FramesStream class.

        Calls relevant classmethod.
        """
        return type(self).class_exts()

    def close(self):
        """
        A method to clean up anything that need to be cleaned up.

        Sub-classes should use super to call up the MRO stack and then
        do any class-specific clean up
        """
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __repr__(self):
        # May be overwritten by subclasses
        return """<Frames>
Frame Shape: {frame_shape!r}
Pixel Datatype: {dtype}""".format(frame_shape=self.frame_shape,
                                  dtype=self.pixel_type)

@Slicerator.from_class
class FramesSequence(FramesStream):
    """Baseclass for wrapping data buckets that have random access.

    Support random access.

    Supports standard slicing and fancy slicing and returns a resliceable
    Slicerator object.

    Must be finite length.

    """
    propagate_attrs = ['frame_shape', 'pixel_type']

    def __getitem__(self, key):
        """__getitem__ is handled by Slicerator. In all pims readers, the data
        returning function is get_frame."""
        return self.get_frame(key)

    def __iter__(self):
        return iter(self[:])

    @property
    def shape(self):
        return (len(self), *self.frame_shape)

    @abstractmethod
    def __len__(self):
        """
        It is obligatory that sub-classes define a length.
        """
        pass

    @abstractmethod
    def get_frame(self, ind):
        """
        Sub classes must over-ride this function for how to get a given
        frame out of the file.  Any data-type specific internal-state
        nonsense should be dealt with in this function.
        """
        pass

    def __repr__(self):
        # May be overwritten by subclasses
        return """<Frames>
Length: {count} frames
Frame Shape: {w} x {h}
Pixel Datatype: {dtype}""".format(w=self.frame_shape[0],
                                  h=self.frame_shape[1],
                                  count=len(self),
                                  dtype=self.pixel_type)


class FrameRewindableStream(FramesStream):
    """
    A base class for holding the common code for
    wrapping data sources that do not rewind easily.
    """
    @abstractmethod
    def rewind(self, j=0):
        """
        Resets the stream to frame j

        j : int
            Frame to rewind the stream to
        """
        pass

    @abstractmethod
    def skip_forward(self, j):
        """
        Skip the stream forward by j frames.

        j : int
           Number of frames to skip
        """
        pass

    @abstractmethod
    def next(self):
        """
        return the next frame in the stream
        """
        pass

    @abstractmethod
    def __len__(self):
        pass

    @abstractproperty
    def current(self):
        """
        The current location in the stream.

        Can be an int if in stream or None if out the end.

        """
        pass

    def __iter__(self):
        self.rewind(0)
        return self

    def __getitem__(self, arg):
        """
        Returns a generator which yields frames
        """
        if isinstance(arg, slice):
            # get value from slice
            start, stop, step = arg.start, arg.stop, arg.step
            # sanitize step
            if step is None:
                step = 1
            if step < 1:
                raise ValueError("step must be positive")
            # make sure the stream is in the right place to start
            if start is None:
                start = 0
            if start < self.current:
                self.rewind(start)
            if start > self.current:
                self.skip_forward(start - self.current)

            # sanity check
            if stop is not None and stop < start:
                raise ValueError("start must be less than stop")
            # special case, we can't just return self, because __iter__ rewinds
            if step == 1 and stop is None:
                # keep going until exhausted
                return (self.next() for _ in itertools.repeat(True))

            return self._step_gen(step, stop)

        elif isinstance(arg, int):
            self.rewind(arg)
            return self.next()
        else:
            raise ValueError("Invalid argument, use either a `slice` or " +
                             "or an `int`. not {t}".format(t=str(type(arg))))

    def _step_gen(self, step, stop):
        """
        Wraps up the logic of stepping forward by step > 1
        """
        while stop is None or self.current < stop:
            yield self.next()
            self.skip_forward(step - 1)
        else:
            raise StopIteration

    def __repr__(self):
        # May be overwritten by subclasses
        return """<Frames>
Length: {count} frames
Frame Shape: {w} x {h}
Pixel Datatype: {dtype}""".format(w=self.frame_shape[0],
                                  h=self.frame_shape[1],
                                  count=len(self),
                                  dtype=self.pixel_type)

def _iter_attr(obj):
    try:
        for ns in [obj] + obj.__class__.mro():
            for attr in ns.__dict__:
                yield ns.__dict__[attr]
    except AttributeError:
        raise StopIteration  # obj has no __dict__


def _transpose(get_frame, expected_axes, desired_axes):
    if list(expected_axes) == list(desired_axes):
        return get_frame
    else:
        transposition = [expected_axes.index(a) for a in desired_axes]
        def get_frame_T(**ind):
            return get_frame(**ind).transpose(transposition)
        return get_frame_T


def _bundle(get_frame, expected_axes, to_iter, sizes, dtype):
    bundled_axes = to_iter + expected_axes
    shape = [sizes[a] for a in bundled_axes]
    iter_shape = shape[:len(to_iter)]
    def get_frame_bundled(**ind):
        result = np.empty(shape, dtype=dtype)
        md_list = []
        for indices in itertools.product(*[range(s) for s in iter_shape]):
            ind.update({n: i for n, i in zip(to_iter, indices)})
            frame = get_frame(**ind)
            result[indices] = frame

            if hasattr(frame, 'metadata'):
                if frame.metadata is not None:
                    md_list.append(frame.metadata)
        # propagate metadata
        if len(md_list) == np.prod(iter_shape):
            metadata = dict()
            keys = md_list[0].keys()
            for k in keys:
                try:
                    metadata[k] = [row[k] for row in md_list]
                except KeyError:
                    # if a field is not present in every frame, ignore it
                    warn('metadata field {} is not propagated')
                else:
                    # if all values are equal, only return one value
                    if metadata[k][1:] == metadata[k][:-1]:
                        metadata[k] = metadata[k][0]
                    else:  # cast into ndarray
                        metadata[k] = np.array(metadata[k])
                        metadata[k].shape = iter_shape
        else:
            metadata = None
        return Frame(result, metadata=metadata)
    return get_frame_bundled, bundled_axes


def _drop(get_frame, expected_axes, to_drop):
    # sort axes in descending order for correct function of np.take
    to_drop_inds = [list(expected_axes).index(a) for a in to_drop]
    indices = np.argsort(to_drop_inds)
    axes = [to_drop_inds[i] for i in reversed(indices)]
    to_drop = [to_drop[i] for i in reversed(indices)]
    result_axes = [a for a in expected_axes if a not in to_drop]

    def get_frame_dropped(**ind):
        result = get_frame(**ind)
        for (ax, name) in zip(axes, to_drop):
            result = np.take(result, ind[name], axis=ax)
        return result
    return get_frame_dropped, result_axes


def _make_get_frame(result_axes, get_frame_dict, sizes, dtype):
    methods = list(get_frame_dict.keys())
    result_axes = [a for a in result_axes]
    result_axes_set = set(result_axes)

    # search for get_frame methods that return the right axes
    for axes in methods:
        if len(set(axes) ^ result_axes_set) == 0:
            # _transpose does nothing when axes == result_axes
            return _transpose(get_frame_dict[axes], axes, result_axes)

    # we need either to drop axes or to iterate over axes:
    # collect some numbers to decide what to do
    arr = [None] * len(methods)
    for i, method in enumerate(methods):
        axes_set = set(method)
        to_iter_set = result_axes_set - axes_set
        to_iter = [x for x in result_axes if x in to_iter_set]  # fix the order
        n_iter = int(np.prod([sizes[ax] for ax in to_iter]))
        to_drop = list(axes_set - result_axes_set)
        n_drop = int(np.prod([sizes[ax] for ax in to_drop]))
        arr[i] = [method, axes_set, to_iter, n_iter, to_drop, n_drop]

    # try to read as less data as possible: try n_drop == 0
    # sort in increasing number of iterations
    arr.sort(key=lambda x: x[3])
    for method, axes_set, to_iter, n_iter, to_drop, n_drop in arr:
        if n_drop > 0:
            continue
        bundled_axes = to_iter + list(method)
        get_frame, after_bundle = _bundle(get_frame_dict[method], method,
                                          to_iter, sizes, dtype)
        return _transpose(get_frame, bundled_axes, result_axes)

    # try to iterate without dropping axes
    # sort in increasing number of dropped frames
    # TODO: sometimes dropping some data is better than having many iterations
    arr.sort(key=lambda x: x[5])
    for method, axes_set, to_iter, n_iter, to_drop, n_drop in arr:
        if n_iter > 0:
            continue
        get_frame, after_drop = _drop(get_frame_dict[method], method, to_drop)
        return _transpose(get_frame, after_drop, result_axes)

    # worst case: all methods have both too many axes and require iteration
    # take lowest number of dropped frames
    # if indecisive, take lowest number of iterations
    arr.sort(key=lambda x: (x[3], x[5]))
    method, axes_set, to_iter, n_iter, to_drop, n_drop = arr[0]

    get_frame, after_drop = _drop(get_frame_dict[method], method, to_drop)
    get_frame, after_bundle = _bundle(get_frame, after_drop, to_iter,
                                      sizes, dtype)
    return _transpose(get_frame, after_bundle, result_axes)


class DefaultCoordsDict(dict):
    """Dictionary that checks whether all keys are in `axes`"""
    def __init__(self, default_coords=None):
        """There is no check done here"""
        super(DefaultCoordsDict, self).__init__()
        self.axes = []

    def __setitem__(self, attr, value):
        if attr not in self.axes:
            raise ValueError("axes %r does not exist" % attr)
        super(DefaultCoordsDict, self).__setitem__(attr, value)

    def update(self, *args, **kwargs):
        # So that update does the check too
        for k, v in dict(*args, **kwargs).items():
            self[k] = v


class FramesSequenceND(FramesSequence):
    """ A base class defining a FramesSequence with an arbitrary number of
    axes. In the context of this reader base class, dimensions like 'x', 'y',
    't' and 'z' will be called axes. Indices along these axes will be called
    coordinates.

    The properties `bundle_axes`, `iter_axes`, and `default_coords` define
    to which coordinates each index points. See below for a description of
    each attribute.

    Subclassed readers only need to define `pixel_type` and `__init__`. At least
    one reader method needs to be registered as such using
    `self._register_get_frame(method, <list of axes>)`.
    In the `__init__`, axes need to be initialized using `_init_axis(name, size)`.
    It is recommended to set default values to `bundle_axes` and `iter_axes`.

    The attributes `__len__`, `get_frame`, and the attributes below are defined
    by this base_class; these should not be changed by derived classes.

    Attributes
    ----------
    axes : list of strings
        List of all available axes
    ndim : int
        Number of image axes
    sizes : dict of int
        Dictionary with all axis sizes
    frame_shape : tuple of int
        Shape of frames that will be returned by get_frame
    iter_axes : iterable of strings
        This determines which axes will be iterated over by the FramesSequence.
        The last element in will iterate fastest. Default [].
    bundle_axes : iterable of strings
        This determines which axes will be bundled into one Frame. The axes in
        the ndarray that is returned by get_frame have the same order as the
        order in this list. Default ['y', 'x'].
    default_coords: dict of int
        When an axis is not present in both iter_axes and bundle_axes, the
        coordinate contained in this dictionary will be used. Default 0 for each.

    Examples
    --------
    >>> class DummyReaderND(FramesSequenceND):
    ...    @property
    ...    def pixel_type(self):
    ...        return 'uint8'
    ...    def __init__(self, shape, **axes):
    ...        super(DummyReaderND, self).__init__()  # properly initialize
    ...        self._init_axis('y', shape[0])
    ...        self._init_axis('x', shape[1])
    ...        for name in axes:
    ...            self._init_axis(name, axes[name])
    ...        self._register_get_frame(self.get_frame_2D, 'yx')
    ...        self.bundle_axes = 'yx'  # set default value
    ...        if 't' in axes:
    ...            self.iter_axes = 't'  # set default value
    ...    def get_frame_2D(self, **ind):
    ...        return np.zeros((self.sizes['y'], self.sizes['x']),
    ...                        dtype=self.pixel_type)

    >>> frames = MDummy((64, 64), t=80, c=2, z=10, m=5)
    >>> frames.bundle_axes = 'czyx'
    >>> frames.iter_axes = 't'
    >>> frames.default_coords['m'] = 3
    >>> frames[5]  # returns Frame at T=5, M=3 with shape (2, 10, 64, 64)
    """
    def __init__(self):
        self._clear_axes()
        self._get_frame_dict = dict()

    def _register_get_frame(self, method, axes):
        axes = tuple([a for a in axes])
        if not hasattr(self, '_get_frame_dict'):
            warn("Please call FramesSequenceND.__init__() at the start of the"
                 "the reader initialization.")
            self._get_frame_dict = dict()
        self._get_frame_dict[axes] = method

    def _clear_axes(self):
        self._sizes = {}
        self._default_coords = DefaultCoordsDict()
        self._iter_axes = []
        self._bundle_axes = ['y', 'x']
        self._get_frame_wrapped = None

    def _init_axis(self, name, size, default=0):
        # check if the axes have been initialized, if not, do it here
        if not hasattr(self, '_sizes'):
            warn("Please call FramesSequenceND.__init__() at the start of the"
                 "the reader initialization.")
            self._clear_axes()
            self._get_frame_dict = dict()
        if name in self._sizes:
            raise ValueError("axis '{}' already exists".format(name))
        self._sizes[name] = int(size)
        self.default_coords.axes = self.axes
        self.default_coords[name] = int(default)

    def __len__(self):
        return int(np.prod([self._sizes[d] for d in self._iter_axes]))

    @property
    def frame_shape(self):
        """ Returns the shape of the frame as returned by get_frame. """
        return tuple([self._sizes[d] for d in self._bundle_axes])

    @property
    def axes(self):
        """ Returns a list of all axes. """
        return [k for k in self._sizes]

    @property
    def ndim(self):
        """ Returns the number of axes. """
        return len(self._sizes)

    @property
    def sizes(self):
        """ Returns a dict of all axis sizes. """
        return self._sizes

    @property
    def bundle_axes(self):
        """ This determines which axes will be bundled into one Frame.
        The ndarray that is returned by get_frame has the same axis order
        as the order of `bundle_axes`.
        """
        return self._bundle_axes[:]  # return a copy

    @bundle_axes.setter
    def bundle_axes(self, value):
        value = list(value)
        invalid = [k for k in value if k not in self._sizes]
        if invalid:
            raise ValueError("axes %r do not exist" % invalid)

        for k in value:
            if k in self._iter_axes:
                del self._iter_axes[self._iter_axes.index(k)]

        self._bundle_axes = value
        if not hasattr(self, '_get_frame_dict'):
            warn("Please call FramesSequenceND.__init__() at the start of the"
                 "the reader initialization.")
            self._get_frame_dict = dict()
        if len(self._get_frame_dict) == 0:
            if hasattr(self, 'get_frame_2D'):
                # include get_frame_2D for backwards compatibility
                self._register_get_frame(self.get_frame_2D, 'yx')
            else:
                raise RuntimeError('No reader methods found. Register a reader '
                                   'method with _register_get_frame')

        # update the get_frame method
        get_frame = _make_get_frame(self._bundle_axes, self._get_frame_dict,
                                    self.sizes, self.pixel_type)
        self._get_frame_wrapped = get_frame

    @property
    def iter_axes(self):
        """ This determines which axes will be iterated over by the
        FramesSequence. The last element will iterate fastest. """
        return self._iter_axes[:]  # return a copy

    @iter_axes.setter
    def iter_axes(self, value):
        value = list(value)
        invalid = [k for k in value if k not in self._sizes]
        if invalid:
            raise ValueError("axes %r do not exist" % invalid)

        for k in value:
            if k in self._bundle_axes:
                del self._bundle_axes[self._bundle_axes.index(k)]

        self._iter_axes = value

    @property
    def default_coords(self):
        """ When a axis is not present in both iter_axes and bundle_axes, the
        coordinate contained in this dictionary will be used. """
        return self._default_coords  # this is a custom dict (DefaultCoordsDict)

    @default_coords.setter
    def default_coords(self, value):
        self._default_coords.update(**value)

    def get_frame(self, i):
        """ Returns a Frame of shape determined by bundle_axes. The index value
        is interpreted according to the iter_axes property. Coordinates not
        present in both iter_axes and bundle_axes will be set to their default
        value (see default_coords). """
        if i > len(self):
            raise IndexError('index out of range')
        if self._get_frame_wrapped is None:
            self.bundle_axes = tuple(self.bundle_axes)  # kick bundle_axes

        # start with the default coordinates
        coords = self.default_coords.copy()

        # list sizes of iteration axes
        iter_sizes = [self._sizes[k] for k in self.iter_axes]
        # list how much i has to increase to get an increase of coordinate n
        iter_cumsizes = np.append(np.cumprod(iter_sizes[::-1])[-2::-1], 1)
        # calculate the coordinates and update the coords dictionary
        iter_coords = (i // iter_cumsizes) % iter_sizes
        coords.update(**{k: v for k, v in zip(self.iter_axes, iter_coords)})

        result = self._get_frame_wrapped(**coords)
        if hasattr(result, 'metadata'):
            metadata = result.metadata
        else:
            metadata = dict()

        metadata_axes = set(self.axes) - set(self.bundle_axes)
        metadata_coords = {ax: coords[ax] for ax in metadata_axes}
        metadata.update(dict(axes=self.bundle_axes, coords=metadata_coords))
        return Frame(result, frame_no=i, metadata=metadata)

    def __repr__(self):
        s = "<FramesSequenceND>\nAxes: {0}\n".format(self.ndim)
        for dim in self._sizes:
            s += "Axis '{0}' size: {1}\n".format(dim, self._sizes[dim])
        s += """Pixel Datatype: {dtype}""".format(dtype=self.pixel_type)
        return s




import uuid
import itertools
import numpy as np
import tempfile
from io import BytesIO
from base64 import b64encode
from contextlib import contextmanager
from fractions import Fraction
import warnings

try:
    from matplotlib.colors import ColorConverter
    import matplotlib as mpl
    import matplotlib.pyplot as plt
except ImportError:
    ColorConverter = None
    mpl = None
    plt = None

try:
    import av
except ImportError:
    av = None

try:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        from moviepy.editor import VideoClip
except ImportError:
    VideoClip = None
except RuntimeError:
    # there is an incompatibility between moviepy 2.3.5 and imageio >= 2.5.0
    VideoClip = None


def export_pyav(sequence, filename, rate=30, bitrate=None,
                width=None, height=None, format=None, codec='mpeg4',
                pixel_format='yuv420p', autoscale=None, quality=None,
                options=None, rate_range=(16, 32)):
    """Export a sequence of images as a standard video file using PyAv.

    N.B. If the quality and detail are insufficient, increase the
    bitrate.

    Parameters
    ----------
    sequence : any iterator or array of array-like images
        The images should have two dimensions plus an
        optional third dimensions representing color.
    filename : string
        name of output file
    rate : integer
        frame rate of output file, 30 by default
    bitrate : integer
        Video bitrate is crudely guessed if None is given.
    width : integer
        By default, set the width of the images.
    height : integer
        By default, set the  height of the images. If width is specified
        and height is not, the height is autoscaled to maintain the aspect
        ratio.
    codec : string
        a valid video encoding, 'mpeg4' by default. Must be supported by the
        container format. Examples are {'mpeg4', 'wmv2', 'libx264', 'rawvideo'}
        Check https://www.ffmpeg.org/ffmpeg-codecs.html#Video-Encoders.
    format : string
        The container format. Guesses from the filename by default.
    pixel_format: string
        Video stream format, 'yuv420p' by default.
        Another possibility is 'bgr24' in combination with the 'rawvideo' codec.
    autoscale : boolean
        Linearly rescale the brightness to use the full gamut of black to
        white values. False by default for uint8 readers, True otherwise.
    quality: number or string, optional
        For 'mpeg4' codec: sets qmin and qmax
        For 'libx264' codec: sets crf. 0 = lossless, 23 = default.
        For 'wmv2' codec: sets fraction of lossless bitrate, 0.01 = default
    options : dictionary, optional
        Dictionary that will be passed to ffmpeg. Avoid using
        {'qscale:v', 'crf', 'pixel_format'}.
    rate_range : tuple of two numbers
        As extreme frame rates have playback issues on many players, by default
        the frame rate is limited between 16 and 32. When the desired frame rate
        is too low, frames will be multiplied an integer number of times. When
        the desired frame rate is too high, frames will be skipped at constant
        intervals.

    """
    if av is None:
        raise("This feature requires PyAV with FFmpeg or libav installed.")

    export_rate = _normalize_framerate(rate, *rate_range)
    sequence = CachedFrameGenerator(sequence, rate, autoscale)

    # pyav is picky with unicode strings
    codec = str(codec)
    if format is not None:
        format = str(format)
    if options is not None:
        for key in options:
            options[str(key)] = str(options[key])
    else:
        options = dict()

    if codec == str('wmv2') and bitrate is None and quality is None:
        quality = 0.01

    if quality is not None:
        if codec == str('libx264'):
            options[str('crf')] = str(quality)
        elif codec == str('wmv2'):
            if bitrate is not None:
                warnings.warn("(wmv) quality is ignored when bitrate is set.")
        elif codec == str('mpeg4'):
            options[str('qmin')] = str(quality)
            options[str('qmax')] = str(quality)
        else:
            raise NotImplemented

    # Maximum allowed timebase is 66535 (at least for mpeg4)
    # see https://github.com/mikeboers/PyAV/issues/242
    export_rate_frac = Fraction(export_rate).limit_denominator(65535)

    output = av.open(str(filename), str('w'), format=format)
    try:
        # from PyAv 6.0, options can be supplied here
        stream = output.add_stream(
            codec, rate=export_rate_frac, options=options
        )
    except TypeError:  # before, we should supply it at .open
        output = av.open(
            str(filename), str('w'), format=format, options=options
        )
        stream = output.add_stream(codec, rate=export_rate_frac)

    stream.pix_fmt = str(pixel_format)

    for frame_no in itertools.count():
        try:
            img = sequence(frame_no / export_rate)
        except IndexError:
            break
        if frame_no == 0:
            # Inspect first frame to set up stream.
            if width is None:
                stream.height = img.shape[0]
                stream.width = img.shape[1]
            else:
                stream.width = width
                stream.height = (height or
                                 width * img.shape[0] // img.shape[1])

            if bitrate is not None:
                stream.bit_rate = int(bitrate)
            elif quality is not None and codec == str('wmv2'):
                bitrate = quality * _estimate_bitrate([stream.height,
                                                       stream.width],
                                                      export_rate)
                stream.bit_rate = int(bitrate)

        # Ensure correct memory layout
        img = img.astype(img.dtype, order='C', copy=False)
        frame = av.VideoFrame.from_ndarray(img, format=str('rgb24'))
        packet = stream.encode(frame)
        if packet is not None:
            output.mux(packet)

    # Finish encoding the stream
    while True:
        try:
            packet = stream.encode()
        except av.AVError:  # End of file raises AVError since after av 0.4
            break
        if packet is None:
            break
        output.mux(packet)

    output.close()


def play(sequence, rate=30, bitrate=None,
         width=None, height=None, autoscale=True):
    """In an IPython notebook, display a sequence of images as
    an embedded video.

    N.B. If the quality and detail are insufficient, increase the
    bit rate.

    Parameters
    ----------
    sequence : any iterator or array of array-like images
        The images should have two dimensions plus an
        optional third dimensions representing color.
    rate : integer
        frame rate of output file, 30 by default
    bitrate : integer
        Video bitrate is crudely guessed if None is given.
    width : integer
        By default, set the width of the images.
    height : integer
        By default, set the  height of the images. If width is specified
        and height is not, the height is autoscaled to maintain the aspect
        ratio.
    autoscale : boolean
        Linearly rescale the brightness to use the full gamut of black to
        white values. If the datatype of the images is not 'uint8', this must
        be set to True, as it is by default.

    """
    try:
        from IPython.display import display
    except ImportError:
        raise ImportError("This feature requires IPython.")
    with tempfile.NamedTemporaryFile(suffix='.webm') as temp:
        export_pyav(sequence, bytes(temp.name), codec='libvpx', rate=rate,
                    width=width, height=height, bitrate=bitrate,
                    format='yuv420p', autoscale=True)
        temp.flush()
        display(repr_video(temp.name, 'x-webm'))


class CachedFrameGenerator(object):
    def __init__(self, sequence, rate, autoscale=None, to_bgr=False):
        self.sequence = sequence
        self._cached_frame_no = None
        self._cache = None
        self.autoscale = autoscale
        self.rate = rate
        self.to_bgr = to_bgr

    def __call__(self, t):
        frame_no = int(t * self.rate)
        if self._cached_frame_no != frame_no:
            self._cached_frame_no = frame_no
            self._cache = _to_rgb_uint8(self.sequence[frame_no], self.autoscale)
        if self.to_bgr:
            return self._cache[:, :, ::-1]
        else:
            return self._cache


def export_moviepy(sequence, filename, rate=30, bitrate=None, width=None,
                   height=None, codec='mpeg4', pixel_format='yuv420p',
                   autoscale=None, quality=None, verbose=True,
                   options=None, rate_range=(16, 32)):
    """Export a sequence of images as a standard video file using MoviePy.

    Parameters
    ----------
    sequence : any iterator or array of array-like images
        The images should have two dimensions plus an
        optional third dimensions representing color.
    filename : string
        name of output file
    rate : integer, optional
        frame rate of output file, 30 by default
        NB: The output frame rate will be limited between `rate_range`
    bitrate : integer or string, optional
        Preferably use the parameter `quality` for controlling the bitrate.
    width : integer, optional
        By default, set the width of the images.
    height : integer, optional
        By default, set the  height of the images. If width is specified
        and height is not, the height is autoscaled to maintain the aspect
        ratio.
    codec : string
        a valid video encoding, 'mpeg4' by default. Must be supported by the
        container format. Examples are {'mpeg4', 'wmv2', 'libx264', 'rawvideo'}
        Check https://www.ffmpeg.org/ffmpeg-codecs.html#Video-Encoders.
    pixel_format: string, optional
        Pixel format, 'yuv420p' by default.
        Another possibility is 'bgr24' in combination with the 'rawvideo' codec.
    quality: number or string, optional
        For 'mpeg4' codec: sets qscale:v. 1 = high quality, 5 = default.
        For 'libx264' codec: sets crf. 0 = lossless, 23 = default.
        For 'wmv2' codec: sets fraction of lossless bitrate, 0.01 = default
    autoscale : boolean, optional
        Linearly rescale the brightness to use the full gamut of black to
        white values. False by default for uint8 readers, True otherwise.
    verbose : boolean, optional
        Determines whether MoviePy will print progress. True by default.
    options : dictionary, optional
        Dictionary of parameters that will be passed to ffmpeg. Avoid using
        {'qscale:v', 'crf', 'pixel_format'}.
    rate_range : tuple of two numbers
        As extreme frame rates have playback issues on many players, by default
        the frame rate is limited between 16 and 32. When the desired frame rate
        is too low, frames will be multiplied an integer number of times. When
        the desired frame rate is too high, frames will be skipped at constant
        intervals.

    See Also
    --------
    http://zulko.github.io/moviepy/ref/VideoClip/VideoClip.html#moviepy.video.VideoClip.VideoClip.write_videofile
    """
    if VideoClip is None:
        raise ImportError('The MoviePy exporter requires moviepy to work.')

    if options is None:
        options = dict()
    ffmpeg_params = []
    for key in options:
        ffmpeg_params.extend(['-{}'.format(key), str(options[key])])

    if rate <= 0:
        raise ValueError
    export_rate = _normalize_framerate(rate, *rate_range)

    clip = VideoClip(CachedFrameGenerator(sequence, rate, autoscale,
                                          to_bgr=(pixel_format == 'bgr24')))
    clip.duration = len(sequence) / rate
    if not (height is None and width is None):
        clip = clip.resize(height=height, width=width)

    if codec == 'wmv2' and bitrate is None and quality is None:
        quality = 0.01

    if quality is not None:
        if codec == 'libx264':
            ffmpeg_params.extend(['-crf', str(quality)])
        elif codec == 'mpeg4':
            ffmpeg_params.extend(['-qscale:v', str(quality)])
        elif codec == 'wmv2':
            if bitrate is not None:
                warnings.warn("(wmv) quality is ignored when bitrate is set.")
            else:
                bitrate = quality * _estimate_bitrate(clip.size, export_rate)
        else:
            raise NotImplemented

    if format is not None:
        ffmpeg_params.extend(['-pixel_format', str(pixel_format)])
    if bitrate is not None:
        bitrate = str(bitrate)

    clip.write_videofile(filename, export_rate, codec, bitrate, audio=False,
                         verbose=verbose, ffmpeg_params=ffmpeg_params)


if av is not None:
    export = export_pyav
elif VideoClip is not None:
    export = export_moviepy
else:
    export = None


def repr_video(fname, mimetype):
    """Load the video in the file `fname`, with given mimetype,
    and display as HTML5 video.
    """
    try:
        from IPython.display import HTML
    except ImportError:
        raise ImportError("This feature requires IPython.")
    video_encoded = open(fname, "rb").read().encode("base64")

    video_tag = """<video controls>
<source alt="test" src="data:video/{0};base64,{1}" type="video/webm">
Use Google Chrome browser.</video>""".format(mimetype, video_encoded)
    return HTML(data=video_tag)


def _scrollable_stack(sequence, width, normed=True):
    # See the public function, scrollable_stack, below.
    # This does all the work, and it returns a string of HTML and JS code,
    # as expected by Frame._repr_html_(). The public function wraps this
    # in IPython.display.HTML for the user.
    from IPython.display import Javascript, HTML, display_png
    from jinja2 import Template

    SCROLL_STACK_JS = Template("""
require(['jquery'], function() {
  if (!(window.PIMS)) {
    var stack_cursors = {};
    window.PIMS = {stack_cursors: {}};
  }
  $('#stack-{{stack_id}}-slice-0').css('display', 'block');
  window.PIMS.stack_cursors['{{stack_id}}'] = 0;
});

require(['jquery'],
$('#image-stack-{{stack_id}}').bind('mousewheel DOMMouseScroll', function(e) {
  var direction;
  var cursor = window.PIMS.stack_cursors['{{stack_id}}'];
  e.preventDefault();
  if (e.type == 'mousewheel') {
    direction = e.originalEvent.wheelDelta < 0;
  }
  else if (e.type == 'DOMMouseScroll') {
    direction = e.originalEvent.detail < 0;
  }
  var delta = direction * 2 - 1;
  if (cursor + delta < 0) {
    return;
  }
  else if (cursor + delta > {{length}} - 1) {
    return;
  }
  $('#stack-{{stack_id}}-slice-' + cursor).css('display', 'none');
  $('#stack-{{stack_id}}-slice-' + (cursor + delta)).css('display', 'block');
  window.PIMS.stack_cursors['{{stack_id}}'] = cursor + delta;
}));""")
    TAG = Template('<img src="data:image/png;base64,{{data}}" '
                   'style="display: none;" '
                   'id="stack-{{stack_id}}-slice-{{i}}" />')
    WRAPPER = Template('<div id="image-stack-{{stack_id}}", style='
                       '"width: {{width}}; float: left; display: inline;">')
    stack_id = uuid.uuid4()  # random unique identifier
    js = SCROLL_STACK_JS.render(length=len(sequence), stack_id=stack_id)
    output = '<script>{0}</script>'.format(js)
    output += WRAPPER.render(width=width, stack_id=stack_id)
    if normed:
        sequence = normalize(np.asarray(sequence))
    for i, s in enumerate(sequence):
        output += TAG.render(
            data=b64encode(_as_png(s, width, normed=False)).decode('utf-8'),
            stack_id=stack_id, i=i)
    output += "</div>"
    return output


def scrollable_stack(sequence, width=512, normed=True):
    """Display a sequence or 3D stack of frames as an interactive image
    that responds to scrolling.

    Parameters
    ----------
    sequence: a 3D Frame (or any array) or an iterable of 2D Frames (or arrays)
    width: integer
        Optional, defaults to 512. The height is auto-scaled.
    normed : Rescale the brightness to fill the gamut. All pixels in the
        stack rescaled uniformly.

    Returns
    -------
    an interactive image, contained in a IPython.display.HTML object
    """
    from IPython.display import HTML
    return HTML(_scrollable_stack(sequence, width=width, normed=normed))


def _as_png(arr, width, normed=True):
    """Create a PNG image buffer from an array."""
    try:
        from PIL import Image
    except ImportError:
        raise ImportError("This feature requires PIL/Pillow.")
    w = width  # for brevity
    h = arr.shape[0] * w // arr.shape[1]
    if normed:
        arr = normalize(arr)
    img = Image.fromarray((arr * 255).astype('uint8')).resize((w, h))
    img_buffer = BytesIO()
    img.save(img_buffer, format='png')
    return img_buffer.getvalue()


def normalize(arr):
    """This normalizes an array to values between 0 and 1.

    Parameters
    ----------
    arr : ndarray

    Returns
    -------
    ndarray of float
        normalized array
    """
    ptp = arr.max() - arr.min()
    # Handle edge case of a flat image.
    if ptp == 0:
        ptp = 1
    scaled_arr = (arr - arr.min()) / ptp
    return scaled_arr


def _to_rgb_uint8(image, autoscale):
    if autoscale is None:
        autoscale = image.dtype != np.uint8

    if autoscale:
        image = (normalize(image) * 255).astype(np.uint8)
    elif image.dtype != np.uint8:
        if np.issubdtype(image.dtype, np.integer):
            max_value = np.iinfo(image.dtype).max
            # sometimes 12-bit images are stored as unsigned 16-bit
            if max_value == 2**16 - 1 and image.max() < 2**12:
                max_value = 2**12 - 1
            image = (image / max_value * 255).astype(np.uint8)
        else:
            image = (image * 255).astype(np.uint8)

    ndim = image.ndim
    shape = image.shape
    if ndim == 3 and shape.count(3) == 1:
        # This is a color image. Ensure that the color axis is axis 2.
        color_axis = shape.index(3)
        image = np.rollaxis(image, color_axis, 3)
    elif image.ndim == 3 and shape.count(4) == 1:
        # This is an RGBA image. Ensure that the color axis is axis 2, and 
        # drop the A values.
        color_axis = shape.index(4)
        image = np.rollaxis(image, color_axis, 3)[:, :, :3]
    elif ndim == 2:
        # Expand into color to satisfy moviepy's expectation
        image = np.repeat(image[:, :, np.newaxis], 3, axis=2)
    else:
        raise ValueError("Images have the wrong shape.")

    return np.asarray(image)


def _estimate_bitrate(shape, frame_rate):
    """Return a bitrate that will guarantee lossless video."""
    # Total Pixels x 8 bits x 3 channels x FPS
    return shape[0] * shape[1] * 8 * 3 * frame_rate


def _normalize_framerate(rate, min_rate=16, max_rate=32):
    """Limits the frame rate between min_rate and max_rate by integer multiples.
    """
    if rate < min_rate:
        factor = min_rate // rate
        if min_rate % rate > 0:
            factor += 1
        return rate * factor

    if rate > max_rate:
        factor = rate // max_rate
        if rate % max_rate > 0:
            factor += 1
        return rate / factor

    return rate


def _monochannel_to_rgb(image, rgb):
    """This converts a greyscale image to an RGB image, using given rgb value.

    Parameters
    ----------
    image : ndarray
        image; there should be no channel axis
    rgb : tuple of uint8
        output color in (r, g, b) format

    Returns
    -------
    ndarray of float
        rgb image, with extra inner dimension of length 3

    """
    image_rgb = normalize(image).reshape(*(image.shape + (1,)))
    image_rgb = image_rgb * np.asarray(rgb).reshape(*((1,)*image.ndim + (3,)))
    return image_rgb


def to_rgb(image, colors=None, normed=True):
    """This converts a greyscale or multichannel image to an RGB image, with
    given channel colors.

    Parameters
    ----------
    image : ndarray
        Multichannel image (channel dimension is first dimension). When first
        dimension is longer than 4, the file is interpreted as a greyscale.
    colors : list of matplotlib.colors
        List of either single letters, or rgb(a) as lists of floats. The sum
        of these lists should equal (1.0, 1.0, 1.0), when clipping needs to
        be avoided.
    normed : bool, optional
        Multichannel images will be downsampled to 8-bit RGB, if normed is
        True. Greyscale images will always give 8-bit RGB.

    Returns
    -------
    ndarray
        RGB image, with inner dimension of length 3. The RGB image is clipped
        so that values lay between 0 and 255. When normed = True (default),
        datatype is np.uint8, else it is float.
    """
    # identify whether the image has a (leading) channel axis
    if colors is None:
        has_channel_axis = image.ndim > 2 and image.shape[0] < 5
    else:
        has_channel_axis = len(colors) == image.shape[0]
    # identify number of channels and resulting shape
    if has_channel_axis:
        channels = image.shape[0]
        shape_rgb = image.shape[1:] + (3,)
    else:
        channels = 1
        shape_rgb = image.shape + (3,)
    if colors is None:
        # pick colors with high RGB luminance
        if channels == 1:    # white
            rgbs = [[255, 255, 255]]
        elif channels == 2:  # green, magenta
            rgbs = [[0, 255, 0], [255, 0, 255]]
        elif channels == 3:  # cyan, green, magenta
            rgbs = [[0, 255, 255], [0, 255, 0], [255, 0, 255]]
        elif channels == 4:  # cyan, green, magenta, red
            rgbs = [[0, 255, 255], [0, 255, 0], [255, 0, 255], [255, 0, 0]]
        else:
            raise IndexError('Not enough color values to build rgb image')
    else:
        # identify rgb values of channels using matplotlib ColorConverter
        if ColorConverter is None:
            raise ImportError('Matplotlib required for conversion to rgb')
        if channels > len(colors):
            raise IndexError('Not enough color values to build rgb image')
        rgbs = (ColorConverter().to_rgba_array(colors)*255).astype('uint8')
        rgbs = rgbs[:channels, :3]

    if has_channel_axis:
        result = np.zeros(shape_rgb)
        for i in range(channels):
            result += _monochannel_to_rgb(image[i], rgbs[i])
    else:
        result = _monochannel_to_rgb(image, rgbs[0])

    result = result.clip(0, 255)

    if normed:
        result = (normalize(result) * 255).astype('uint8')

    return result


@contextmanager
def _fig_size_cntx(fig, fig_size_inches, tight_layout):
    """Resize a figure in a context

    Parameters
    ----------
    fig : matplotlib.figure.Figure
        The figure to resize
    fig_size_inches : tuple
        The (height, width) to use in the context. If None, the size
        is not changed
    tight_layout : boolean
        When True, tight layout is used.
    """
    orig_size = fig.get_size_inches()
    orig_layout = fig.get_tight_layout()
    if fig_size_inches is not None:
        fig.set_size_inches(*fig_size_inches)
    fig.set_tight_layout(tight_layout)
    if tight_layout:
        rc_params = {'savefig.bbox': 'tight'}
    else:
        rc_params = {'savefig.bbox': 'standard'}
    try:
        with plt.rc_context(rc_params):
            yield fig
    finally:
        fig.set_size_inches(*orig_size)
        fig.set_tight_layout(orig_layout)


def plot_to_frame(fig, width=512, close_fig=False, fig_size_inches=None,
                  bbox_inches=None):
    """ Renders a matplotlib figure or axes object into a numpy array
    containing RGBA data of the rendered image.

    Parameters
    ----------
    fig : matplotlib Figure or Axes object
    width : integer
        The width of the resulting frame, in pixels
    close_fig : boolean
        When True, the figure will be closed after plotting
    fig_size_inches : tuple
        The figure (height, width) in inches. If None, the size is not changed.
    bbox_inches : {None, 'standard', 'tight'}
        When 'tight', tight layout is used.

    Returns
    -------
    pims.Frame object containing RGBA values (dtype uint8)
    """
    if mpl is None:
        raise ImportError("This feature requires matplotlib.")
    from pims import Frame
    if isinstance(fig, mpl.axes.Axes):
        fig = fig.figure
    if fig_size_inches is not None:
        if fig_size_inches[0] == 0 or fig_size_inches[1] == 0:
            raise ValueError('Figure size cannot be zero.')
    if bbox_inches is None:
        tight_layout = fig.get_tight_layout()
    elif str(bbox_inches) == 'standard':
        tight_layout = False
    elif str(bbox_inches) == 'tight':
        tight_layout = True
    else:
        raise ValueError("bbox_inches must be in {None, 'standard', 'tight'}")

    buf = BytesIO()
    with _fig_size_cntx(fig, fig_size_inches, tight_layout) as fig:
        width_in, height_in = fig.get_size_inches()
        dpi = width / width_in
        if tight_layout:
            # slower, but allows tight layout
            fig.savefig(buf, format='png', dpi=dpi)
            buf.seek(0)
            image = plt.imread(buf)
        else:
            # faster, but only possible without tight layout
            fig.savefig(buf, format='rgba', dpi=dpi)
            buf.seek(0)
            buf_shape = (int(height_in * dpi), int(width_in * dpi), 4)
            image = np.frombuffer(buf.read(),
                                  dtype='uint8').reshape(*buf_shape)

    if close_fig:
        plt.close(fig)
    return Frame(image)


def plots_to_frame(figures, width=512, close_fig=False, fig_size_inches=None,
                   bbox_inches=None):
    """ Renders an iterable of matplotlib figures or axes objects into a
    pims Frame object, that will be displayed as scrollable stack in IPython.

    Parameters
    ----------
    figures : iterable of matplotlib Figure or Axes objects
    width : integer
        The width of the resulting frame, in pixels
    close_fig : boolean
        When True, the figure will be closed after plotting
    fig_size_inches : tuple
        The figure (height, width) in inches. If None, the size is not changed.
    bbox_inches : {'tight', None}
        When 'tight', tight layout is used.

    Returns
    -------
    pims.Frame object containing a stack of RGBA values (dtype uint8)
    """
    if mpl is None:
        raise ImportError("This feature requires matplotlib.")
    from pims import Frame
    if isinstance(figures, mpl.axes.Axes) or \
       isinstance(figures, mpl.figure.Figure):
        raise ValueError('Use plot_to_frame for single figures, or supply '
                         'an iterable of figures to plots_to_frame.')

    width = int(width)
    h = None

    frames = []
    for n, fig in enumerate(figures):
        im = plot_to_frame(fig, width, close_fig, fig_size_inches, bbox_inches)
        if h is None:
            h = im.shape[0]
        else:
            # make the image the same size as the first image
            if im.shape[0] != h:
                im = np.pad(im[:h], ((0, max(0, h - im.shape[0])), (0, 0),
                                     (0, 0)), mode=str('constant'))
        frames.append(im)

    return Frame(np.array(frames))



import os
import warnings
import numpy as np

from .frame import Frame
from .base_frames import FramesSequence


class Spec(object):
    """SPE file specification data

    Tuples of (offset, datatype, count), where offset is the offset in the SPE
    file and datatype is the datatype as used in `numpy.fromfile`()

    `data_start` is the offset of actual image data.

    `dtypes` translates SPE datatypes (0...4) to numpy ones, e. g. dtypes[0]
    is dtype("<f") (which is np.float32).

    `controllers` maps the `type` metadata to a human readable name

    `readout_modes` maps the `readoutMode` metadata to something human readable
    although this may not be accurate since there is next to no documentation
    to be found.
    """
    metadata = {
        #essential information
        "datatype": (108, "<h"), #dtypes
        "xdim": (42, "<H"),
        "ydim": (656, "<H"),
        "NumFrames": (1446, "<i"),

        #ROI information
        "NumROIsInExperiment": (1488, "<h"),
        "NumROI": (1510, "<h"),
        "ROIs": (1512, np.dtype([("startx", "<H"),
                                 ("endx", "<H"),
                                 ("groupx", "<H"),
                                 ("starty", "<H"),
                                 ("endy", "<H"),
                                 ("groupy", "<H")]), 10),

        #chip-related sizes
        "xDimDet": (6, "<H"),
        "yDimDet": (18, "<H"),
        "VChipXdim": (14, "<h"),
        "VChipYdim": (16, "<h"),

        #other stuff
        "ControllerVersion": (0, "<h"),
        "LogicOutput": (2, "<h"),
        "AmpHiCapLowNoise": (4, "<H"), #enum?
        "mode": (8, "<h"), #enum?
        "exp_sec": (10, "<f"),
        "date": (20, "<10S"),
        "DetTemperature": (36, "<f"),
        "DetType": (40, "<h"),
        "stdiode": (44, "<h"),
        "DelayTime": (46, "<f"),
        "ShutterControl": (50, "<H"), #normal, disabled open, disabled closed
                                      #but which one is which?
        "AbsorbLive": (52, "<h"), #bool?
        "AbsorbMode": (54, "<H"),
        "CanDoVirtualChipFlag": (56, "<h"), #bool?
        "ThresholdMinLive": (58, "<h"), #bool?
        "ThresholdMinVal": (60, "<f"),
        "ThresholdMinLive": (64, "<h"), #bool?
        "ThresholdMinVal": (66, "<f"),
        "ExperimentTimeLocal": (172, "<7S"),
        "ExperimentTimeUTC": (179, "<7S"),
        "ADCoffset": (188, "<H"),
        "ADCrate": (190, "<H"),
        "ADCtype": (192, "<H"),
        "ADCresolution": (194, "<H"),
        "ADCbitAdjust": (196, "<H"),
        "gain": (198, "<H"),
        "comments": (200, "<80S", 5),
        "geometric": (600, "<H"), #flags
        "swversion": (688, "<16S"),
        "spare4": (742, "<436S"),
        "XPrePixels": (98, "<h"),
        "XPostPixels": (100, "<h"),
        "YPrePixels": (102, "<h"),
        "YPostPixels": (104, "<h"),
        "ReadoutTime": (672, "<f"),
        "type": (704, "<h"), #controllers
        "clkspd_us": (1428, "<f"),
        "readoutMode": (1480, "<H"), #readout_modes
        "WindowSize": (1482, "<H"),
        "file_header_ver": (1992, "<f")
    }

    data_start = 4100

    dtypes = [np.dtype("<f"), np.dtype("<i"), np.dtype("<h"),
              np.dtype("<H"), np.dtype("<I")]

    controllers = [
        "new120 (Type II)", "old120 (Type I)", "ST130", "ST121", "ST138",
        "DC131 (PentaMax)", "ST133 (MicroMax/Roper)", "ST135 (GPIB)", "VTCCD",
        "ST116 (GPIB)", "OMA3 (GPIB)", "OMA4"
    ]

    #This was gathered from random places on the internet and own experiments
    #with the camera. May not be accurate.
    readout_modes = ["full frame", "frame transfer", "kinetics"]

    #do not decode the following metadata keys into strings, but leave them
    #as byte arrays
    no_decode = ["spare4"]


class SpeStack(FramesSequence):
    """Read image data from SPE files

    Attributes
    ----------
    default_char_encoding : string
        Default character encoding used to decode metadata strings. This is a
        class attribute. By setting `SpeStack.default_char_encoding =
        "my_encoding"`, "my_encoding" will be used as a default in all SpeStack
        instances thereafter, unless a different one is explicitly passed to
        the constructor. Defaults to "latin1".
    metadata : dict
        Contains additional metadata.
    """
    default_char_encoding = "latin1"

    @classmethod
    def class_exts(cls):
        return {"spe"} | super(SpeStack, cls).class_exts()

    def __init__(self, filename, char_encoding=None, check_filesize=True):
        """Create an iterable object that returns image data as numpy arrays

        Arguments
        ---------
        filename : string
            Name of the SPE file
        char_encoding : str or None, optional
            Specifies what character encoding is used to decode metatdata
            strings. If None, use the `default_char_encoding` class attribute.
            Defaults to None.
        check_filesize : bool, optional
            The number of frames in an SPE file should be recorded in the
            file's header. Some software fails to do so correctly. If
            `check_filesize` is `True`, calculate the number of frames from
            the file size. A warning is emitted if this doesn't match the
            number of frames from the file header. Defaults to True.
        """
        self._filename = filename
        self._file = open(filename, "rb")
        self._char_encoding = (char_encoding if char_encoding is not None
                               else self.default_char_encoding)

        ### Read metadata ###
        self.metadata = {}
        #Decode each string from the numpy array read by np.fromfile
        decode = np.vectorize(lambda x: x.decode(self._char_encoding))

        for name, sp in Spec.metadata.items():
            self._file.seek(sp[0])
            cnt = (1 if len(sp) < 3 else sp[2])
            v = np.fromfile(self._file, dtype=sp[1], count=cnt)
            if v.dtype.kind == "S" and name not in Spec.no_decode:
                #silently ignore string decoding failures
                try:
                    v = decode(v)
                except:
                    pass
            if cnt == 1:
                #for convenience, if the array contains only one single entry,
                #return this entry itself.
                v = v.item()
            self.metadata[name] = v

        ### Some metadata is "special", deal with it
        #Determine data type
        self._dtype = Spec.dtypes[self.metadata.pop("datatype")]

        #movie dimensions
        self._width = self.metadata.pop("xdim")
        self._height = self.metadata.pop("ydim")
        self._len = self.metadata.pop("NumFrames")

        if check_filesize:
            # Some software writes incorrecet `NumFrames` metadata
            # Use the file size to determine the number of frames
            fsz = os.path.getsize(filename)
            l = fsz - Spec.data_start
            l //= self._width * self._height * self._dtype.itemsize
            if l != self._len:
                warnings.warn("Number of frames according to file header "
                              "does not match the size of file " +
                              filename + ".")
                self._len = min(l, self._len)

        #The number of ROIs is given in the SPE file. Only return as many
        #ROIs as given
        num_rois = self.metadata.pop("NumROI", None)
        num_rois = (1 if num_rois < 1 else num_rois)
        self.metadata["ROIs"] = self.metadata["ROIs"][:num_rois]

        #chip sizes
        self.metadata["ChipSize"] = (self.metadata.pop("xDimDet", None),
                                     self.metadata.pop("yDimDet", None))
        self.metadata["VirtChipSize"] = (self.metadata.pop("VChipXdim", None),
                                         self.metadata.pop("VChipYdim", None))

        #geometric operations
        g = []
        f = self.metadata.pop("geometric", 0)
        if f & 1:
            g.append("rotate")
        if f & 2:
            g.append("reverse")
        if f & 4:
            g.append("flip")
        self.metadata["geometric"] = g

        #Make some additional information more human-readable
        t = self.metadata["type"]
        if 1 <= t <= len(Spec.controllers):
            self.metadata["type"] = Spec.controllers[t - 1]
        else:
            self.metadata.pop("type", None)
        m = self.metadata["readoutMode"]
        if 1 <= m <= len(Spec.readout_modes):
            self.metadata["readoutMode"] = Spec.readout_modes[m - 1]
        else:
            self.metadata.pop("readoutMode", None)

    @property
    def frame_shape(self):
        return self._height, self._width

    def __len__(self):
        return self._len

    def get_frame(self, j):
        if j >= self._len:
            raise ValueError("Frame number {} out of range.".format(j))
        self._file.seek(Spec.data_start
                        + j*self._width*self._height*self.pixel_type.itemsize)
        data = np.fromfile(self._file, dtype=self.pixel_type,
                           count=self._width*self._height)
        return Frame(data.reshape(self._height, self._width),
                     frame_no=j, metadata=self.metadata)

    def close(self):
        """Clean up and close file"""
        super(SpeStack, self).close()
        self._file.close()

    @property
    def pixel_type(self):
        return self._dtype

    def __repr__(self):
        return """<Frames>
Source: {filename}
Length: {count} frames
Frame Shape: {w} x {h}
Pixel Datatype: {dtype}""".format(w=self._width,
                                  h=self._height,
                                  count=self._len,
                                  filename=self._filename,
                                  dtype=self._dtype)



import numpy as np

from pims.base_frames import FramesSequence
from pims.frame import Frame


try:
    import av
except ImportError:
    av = None


def available():
    return av is not None


def _next_video_packet(container_iter):
    for packet in container_iter:
        if packet.stream.type == 'video':
            decoded = packet.decode()
            if len(decoded) > 0:
                return decoded

    raise ValueError("Could not find any video packets.")


class WrapPyAvFrame(object):
    def __init__(self, frame, frame_no, metadata=None):
        self.frame_no = frame_no
        self.arr = None
        self.metadata = metadata

        # makes a copy of the frame so that ffmpeg does not reuse the buffer
        # by converting already to rgb24. rgb24 movies actually are converted
        # twice. don't know how to just copy! But the operations are fast.
        if frame.format.name == 'rgb24':
            frame = frame.reformat(format="bgr24")
        self.frame = frame.reformat(format="rgb24")

    def to_frame(self):
        if self.arr is None:
            self.arr = Frame(self.frame.to_ndarray(format='rgb24'),
                             frame_no=self.frame_no, metadata=self.metadata)
        return self.arr


def _gen_frames(demuxer, time_base, frame_rate=1., first_pts=0):
    for packet in demuxer:
        for frame in packet.decode():
            # learn timestamp
            for timestamp in (frame.pts, packet.pts, frame.dts, packet.dts):
                if timestamp is not None:
                    break
            else:
                raise IOError(
                    "Unable to read video: frames contain no timestamps. "
                    "Please use PyAVReaderIndexed.")
            t = (timestamp - first_pts) * time_base
            i = int(round(t * frame_rate))
            yield WrapPyAvFrame(frame, frame_no=i,
                                metadata=dict(timestamp=timestamp, t=float(t)))


class PyAVReaderTimed(FramesSequence):
    """Read images from a video file via a direct FFmpeg/AVbin interface.

    The frames are indexed according to their 'timestamp', starting at 0 at the
    timestamp of the first non-empty frame. Missing frames are filled in with
    empty frames. The number of frames in the video is estimated from the
    movie duration and the average frame rate.

    Parameters
    ----------
    filename : string
    cache_size : integer, optional
        the number of frames that are kept in memory. Default 16.
    fast_forward_thresh : integer, optional
        the reader will proceed through the frames if forwarding below this
        number. If forwarding above this number, it will use seek(). Default 32.
    stream_index : integer, optional
        the index of the video stream inside the file. rarely other than 0.

    Examples
    --------
    >>> video = PyAVVideoReader('video.avi')  # or .mov, etc.
    >>> video[0] # Show the first frame.
    >>> video[-1] # Show the last frame.
    >>> video[1][0:10, 0:10] # Show one corner of the second frame.

    >>> for frame in video[:]:
    ...    # Do something with every frame.

    >>> for frame in video[10:20]:
    ...    # Do something with frames 10-20.

    >>> for frame in video[[5, 7, 13]]:
    ...    # Do something with frames 5, 7, and 13.

    >>> frame_count = len(video) # Number of frames in video
    >>> frame_shape = video.frame_shape # Pixel dimensions of video
    """
    class_priority = 9
    @classmethod
    def class_exts(cls):
        return {'mov', 'avi', 'mp4'} | super(PyAVReaderTimed, cls).class_exts()

    def __init__(self, file, cache_size=16, fast_forward_thresh=32,
                 stream_index=0, format=None):
        if not hasattr(file, 'read'):
            file = str(file)
        self.file = file
        self.format = format
        self._container = av.open(self.file, format=self.format)

        if len(self._container.streams.video) == 0:
            raise IOError("No valid video stream found in {}".format(file))

        self._stream = self._container.streams.video[stream_index]

        try:
            self._duration = self._stream.duration * self._stream.time_base
        except TypeError:
            self._duration = self._container.duration / av.time_base

        self._frame_rate = self._stream.average_rate
        if self.duration <= 0 or len(self) <= 0:
            raise IOError("Video stream {} in {} has zero length.".format(stream_index, file))

        self._cache = [None] * cache_size
        self._fast_forward_thresh = fast_forward_thresh

        demuxer = self._container.demux(self._stream)

        # obtain first frame to get first time point
        # also tests for the presence of timestamps
        frame = next(_gen_frames(demuxer, self._stream.time_base))
        self._first_pts = frame.metadata['timestamp']

        frame = WrapPyAvFrame(frame.frame, 0, frame.metadata)
        self._cache[0] = frame
        self._frame_shape = (self._stream.height, self._stream.width, 3)
        self._last_frame = 0

        self._reset_demuxer()

    def __len__(self):
        return int(self._duration * self._frame_rate)

    def _reset_demuxer(self):
        demuxer = self._container.demux(self._stream)
        self._frame_generator = _gen_frames(demuxer, self._stream.time_base,
                                            self._frame_rate, self._first_pts)

    @property
    def duration(self):
        """The video duration in seconds."""
        return float(self._duration)

    @property
    def frame_shape(self):
        return self._frame_shape

    @property
    def frame_rate(self):
        return float(self._frame_rate)

    def get_frame(self, i):
        cached_frame = self._cache[i % len(self._cache)]
        if cached_frame is None:
            cached_i = -1
        else:
            cached_i = cached_frame.frame_no

        # return directly if the frame is in cache
        if cached_i == i:
            return cached_frame.to_frame()

        # check if we will have to seek to the frame
        if self._last_frame >= i or \
            self._last_frame < i - self._fast_forward_thresh:
            frame = self.seek(i)

            # return directly if the seek was perfect (happens rarely)
            if frame is not None:
                if frame.frame_no == i:
                    return frame.to_frame()

        # proceed through the frames
        result = None
        for frame in self._frame_generator:
            # first cache the frame
            self._cache[frame.frame_no % len(self._cache)] = frame
            self._last_frame = frame.frame_no

            if frame.frame_no < i:
                continue  # go on towards the frame
            elif frame.frame_no == i:
                result = frame
                break
            else:  # the frame was not inside the reader
                break
        else:
            # always restart the frame generator when it ends
            self._reset_demuxer()

        if result is None:
            # the requested frame actually does not exist. Can occur due to
            # a bad file, or due to inaccuracy of reader length __len__.
            # find it in the cache
            for other_i in range(i - 1, i - len(self._cache), -1):
                result = self._cache[other_i % len(self._cache)]
                if result is None:
                    continue
                if result.frame_no < i:
                    break
            else:  # cache is empty: return an empty frame
                return Frame(np.zeros(self.frame_shape, dtype=self.pixel_type),
                             frame_no=i)

        return result.to_frame()

    def seek(self, i):
        """Seek to a frame before i and return the first frame."""
        # flush the cache
        self._cache = [None] * len(self._cache)
        # the ffmpeg decode cache is flushed automatically

        timestamp = int(i / (self._frame_rate * self._stream.time_base))
        self._stream.container.seek(timestamp + self._first_pts)

        # check the first frame
        try:
            frame = next(self._frame_generator)
        except StopIteration:
            self._reset_demuxer()
            try:
                frame = next(self._frame_generator)
            except StopIteration:
                return None

        if i == 0:  # security measure to avoid infinite recursion
            return frame

        if frame.frame_no > i:
            # recurse with an additional offset of 16 frames
            return self.seek(i - 16)

        # add the frame to the cache if succesful
        self._cache[frame.frame_no % len(self._cache)] = frame
        self._last_frame = frame.frame_no
        return frame

    @property
    def pixel_type(self):
        return np.uint8

    def __repr__(self):
        # May be overwritten by subclasses
        return """<Frames>
Format: {format}
Source: {filename}
Duration: {duration:.3f} seconds
Frame rate: {frame_rate:.3f} fps
Length: {count} frames
Frame Shape: {frame_shape!r}
""".format(frame_shape=self.frame_shape,
           format=self._stream.long_name,
           duration=self.duration,
           frame_rate=self.frame_rate,
           count=len(self),
           filename=self.file)


class PyAVReaderIndexed(FramesSequence):
    """Read images from the frames of a standard video file into an
    iterable object that returns images as numpy arrays.

    Parameters
    ----------
    filename : string

    Examples
    --------
    >>> video = Video('video.avi')  # or .mov, etc.
    >>> imshow(video[0]) # Show the first frame.
    >>> imshow(video[-1]) # Show the last frame.
    >>> imshow(video[1][0:10, 0:10]) # Show one corner of the second frame.

    >>> for frame in video[:]:
    ...    # Do something with every frame.

    >>> for frame in video[10:20]:
    ...    # Do something with frames 10-20.

    >>> for frame in video[[5, 7, 13]]:
    ...    # Do something with frames 5, 7, and 13.

    >>> frame_count = len(video) # Number of frames in video
    >>> frame_shape = video.frame_shape # Pixel dimensions of video
    """
    class_priority = 8

    @classmethod
    def class_exts(cls):
        return {'mov', 'avi',
                'mp4'} | super(PyAVReaderIndexed, cls).class_exts()

    def __init__(self, file, toc=None, format=None):
        if not hasattr(file, 'read'):
            file = str(file)
        self.file = file
        self.format = format
        self._container = None

        with av.open(self.file, format=self.format) as container:
            stream = [s for s in container.streams if s.type == 'video'][0]

            # Build a toc
            if toc is None:
                packet_lengths = []
                packet_ts = []
                for packet in container.demux(stream):
                    if packet.stream.type == 'video':
                        decoded = packet.decode()
                        if len(decoded) > 0:
                            packet_lengths.append(len(decoded))
                            packet_ts.append(decoded[0].pts)
                self._toc = {
                    'lengths': packet_lengths,
                    'ts': packet_ts,
                }
            else:
                self._toc = toc

            self._toc_cumsum = np.cumsum(self.toc['lengths'])
            self._len = self._toc_cumsum[-1]

            # PyAV always returns frames in color, and we make that
            # assumption in get_frame() later below, so 3 is hardcoded here:
            self._im_sz = stream.height, stream.width, 3
            self._time_base = stream.time_base

        self._load_fresh_file()

    def _load_fresh_file(self):
        if self._container is not None:
            self._container.close()

        if hasattr(self.file, 'seek'):
            self.file.seek(0)

        self._container = av.open(self.file, format=self.format)
        demux = self._container.demux(self._video_stream)
        self._current_packet = _next_video_packet(demux)
        self._current_packet_no = 0

    @property
    def _video_stream(self):
        return [s for s in self._container.streams if s.type == 'video'][0]

    def __len__(self):
        return self._len

    def __del__(self):
        self._container.close()

    @property
    def frame_shape(self):
        return self._im_sz

    @property
    def toc(self):
        return self._toc

    def get_frame(self, j):
        # Find the packet this frame is in.
        packet_no = self._toc_cumsum.searchsorted(j, side='right')
        self._seek_packet(packet_no)
        # Find the location of the frame within the packet.
        if packet_no == 0:
            loc = j
        else:
            loc = j - self._toc_cumsum[packet_no - 1]
        frame = self._current_packet[loc]  # av.VideoFrame

        return Frame(frame.to_ndarray(format='rgb24'), frame_no=j)

    def _seek_packet(self, packet_no):
        """Advance through the container generator until we get the packet
        we want. Store that packet in selfpp._current_packet."""
        packet_ts = self.toc['ts'][packet_no]
        # Only seek when needed.
        if packet_no == self._current_packet_no:
            return
        elif (packet_no < self._current_packet_no
                or packet_no > self._current_packet_no + 1):
            self._container.seek(packet_ts, stream=self._video_stream)

        demux = self._container.demux(self._video_stream)
        self._current_packet = _next_video_packet(demux)
        while self._current_packet[0].pts < packet_ts:
            self._current_packet = _next_video_packet(demux)

        self._current_packet_no = packet_no

    @property
    def pixel_type(self):
        # No need to detect dtype: PyAV always returns uint8.
        return np.uint8

    def __repr__(self):
        # May be overwritten by subclasses
        return """<Frames>
Source: {filename}
Length: {count} frames
Frame Shape: {frame_shape!r}
""".format(frame_shape=self.frame_shape,
           count=len(self),
           filename=self.file)



"""Reader for Norpix .seq files

Author: Nathan C. Keim
Based heavily on cine.py by Kleckner and Caswell
"""

from pims.frame import Frame
from pims.base_frames import FramesSequence, index_attr
from pims.utils.misc import FileLocker
import os, struct, itertools
from warnings import warn
import datetime
import numpy as np
from threading import Lock

__all__ = ['NorpixSeq',]

DWORD = 'L'
LONG = 'l'
DOUBLE = 'd'
USHORT = 'H'

HEADER_FIELDS = [
    ('magic', DWORD),
    ('name', '24s'),
    ('version', LONG),
    ('header_size', LONG),
    ('description', '512s'),
    ('width', DWORD),
    ('height', DWORD),
    ('bit_depth', DWORD),
    ('bit_depth_real', DWORD),
    ('image_size_bytes', DWORD),
    ('image_format', DWORD),
    ('allocated_frames', DWORD),
    ('origin', DWORD),
    ('true_image_size', DWORD),
    ('suggested_frame_rate', DOUBLE),
    ('description_format', LONG),
    ('reference_frame', DWORD),
    ('fixed_size', DWORD),
    ('flags', DWORD),
    ('bayer_pattern', LONG),
    ('time_offset_us', LONG),
    ('extended_header_size', LONG),
    ('compression_format', DWORD),
    ('reference_time_s', LONG),
    ('reference_time_ms', USHORT),
    ('reference_time_us', USHORT)
    # More header values not implemented
]


class NorpixSeq(FramesSequence):
    """Read Norpix sequence (.seq) files

    This is the native format of StreamPix software, owned by NorPix Inc.
    The format is described in the StreamPix documentation.

    Currently only supports uncompressed files (with the same number
    of bytes for each frame).

    Only unsigned 8, 16 or 32 bit monochrome files are directly supported.
    Other color and monochrome pixel formats can be handled by setting
    the as_raw flag which results in frames with unsigned 8-bit data type.
    and the correct number of rows.  (If a raw frame size in bytes is not
    evenly divisible by the number of rows, a 1-D array is returned).

    Nominally thread-safe.

    Parameters
    ----------
    filename : string
        Path to the .seq file
    as_raw : boolean, optional
        Required for non-monochrome frames.
        Images will be returned as an ndarray of bytes.
        2-dimensional if the image height evenly divides the bytes per image,
        1-dimensional otherwise.
    """
    @classmethod
    def class_exts(cls):
        return {'seq'} | super(NorpixSeq, cls).class_exts()

    propagate_attrs = ['frame_shape', 'pixel_type', 'get_time',
                       'get_time_float', 'filename', 'width', 'height',
                       'frame_rate']

    def __init__(self, filename, as_raw=False):
        super(NorpixSeq, self).__init__()
        self._file = open(filename, 'rb')
        self._filename = filename

        self.header_dict = self._read_header(HEADER_FIELDS)

        if self.header_dict['magic'] != 0xFEED:
            raise IOError('The format of this .seq file is unrecognized')
        if self.header_dict['compression_format'] != 0:
            raise IOError('Only uncompressed images are supported in .seq files')
        if self.header_dict['image_format'] != 100 and not as_raw:
            raise IOError('Non-monochrome images are only supported as_raw in .seq files')

        # File-level metadata
        if self.header_dict['version'] >= 5:  # StreamPix version 6
            self._image_offset = 8192
            # Timestamp = 4-byte unsigned long + 2-byte unsigned short (ms)
            #   + 2-byte unsigned short (us)
            self._timestamp_struct = struct.Struct('<LHH')
            self._timestamp_micro = True
        else:  # Older versions
            self._image_offset = 1024
            self._timestamp_struct = struct.Struct('<LH')
            self._timestamp_micro = False
        self._image_block_size = self.header_dict['true_image_size']
        self._filesize = os.stat(self._filename).st_size
        self._image_count = int((self._filesize - self._image_offset) /
                                self._image_block_size)

        # Image metadata
        self._width = self.header_dict['width']
        self._height = self.header_dict['height']
        self._image_bytes = self.header_dict['image_size_bytes']
        if as_raw:
            self._pixel_count = self._image_bytes
            if self._pixel_count % self._height == 0:
                self._shape = (self._height, int(self._pixel_count / self._height))
            else:
                self._shape = (self._image_bytes,)
            self._dtype = 'uint8'
        else:
            try:
                self._pixel_count = self._width * self._height
                dtype_native = 'uint%i' % self.header_dict['bit_depth']
                self._dtype = np.dtype(dtype_native)
                self._shape = (self._height, self._width)
            except TypeError as e:
                raise IOError(self._dtype + " pixels not supported; use as_raw and convert")

        # Public metadata
        self.metadata = {k: self.header_dict[k] for k in
                         ('description', 'bit_depth_real', 'origin',
                          'suggested_frame_rate', 'width', 'height')}
        self.metadata['gamut'] = 2**self.metadata['bit_depth_real'] - 1

        self._file_lock = Lock()

    def _read_header(self, fields, offset=0):
        self._file.seek(offset)
        tmp = dict()
        for name, format in fields:
            val = self._unpack(format)
            tmp[name] = val

        return tmp

    def _unpack(self, fs, offset=None):
        if offset is not None:
            self._file.seek(offset)
        s = struct.Struct('<' + fs)
        vals = s.unpack(self._file.read(s.size))
        if len(vals) == 1:
            return vals[0]
        else:
            return vals

    def _verify_frame_no(self, i):
        if int(i) != i:
            raise ValueError("Frame numbers can only be integers")
        if i >= self._image_count or i < 0:
            raise ValueError("Frame number is out of range: " + str(i))

    def set_process_func(self, process_func):
        # Expose the _validate_process_func for use after the header is
        # available
        self._validate_process_func(process_func)

    def get_frame(self, i):
        self._verify_frame_no(i)
        with FileLocker(self._file_lock):
            self._file.seek(self._image_offset + self._image_block_size * i)
            imdata = np.fromfile(self._file, self.pixel_type, self._pixel_count
                                 ).reshape(self._shape)
            # Timestamp immediately follows
            tfloat, ts = self._read_timestamp()
            md = {'time': ts, 'time_float': tfloat,
                  'gamut': self.metadata['gamut']}
            return Frame(imdata, frame_no=i, metadata=md)

    def _read_timestamp(self):
        """Read a timestamp at the current position in the file.

        Returns a floating-point representation in seconds, and a datetime instance.
        """
        if self._timestamp_micro:
            tsecs, tms, tus = self._timestamp_struct.unpack(self._file.read(8))
            tfloat = tsecs + float(tms) / 1000. + float(tus) / 1.0e6
        else:
            tsecs, tms = self._timestamp_struct.unpack(self._file.read(6))
            tfloat = tsecs + float(tms) / 1000.
        return tfloat, datetime.datetime.fromtimestamp(tfloat)

    def _get_time(self, i):
        """Call _read_timestamp() for a given frame."""
        self._verify_frame_no(i)
        with FileLocker(self._file_lock):
            self._file.seek(self._image_offset + self._image_block_size * i
                            + self._image_bytes)
            return self._read_timestamp()

    @index_attr
    def get_time(self, i):
        """Return the time of frame i as a datetime instance.

        Calling this function in a different timezone than where the movie
        was recorded will result in an offset. The .seq format does not
        store UTC or timezone information.
        """
        return self._get_time(i)[1]

    @index_attr
    def get_time_float(self, i):
        """Return the time of frame i as a floating-point number of seconds."""
        return self._get_time(i)[0]

    def dump_times_float(self):
        """Return all frame times in file, as an array of floating-point numbers."""
        return np.array([self._get_time(i)[0] for i in range(len(self))])

    @property
    def filename(self):
        return self._filename

    @property
    def pixel_type(self):
        return self._dtype

    @property
    def width(self):
        return self.metadata['width']

    @property
    def height(self):
        return self.metadata['height']

    @property
    def frame_shape(self):
        return (self.metadata['height'], self.metadata['width'])

    @property
    def frame_rate(self):
        return self.metadata['suggested_frame_rate']

    def __len__(self):
        return self._image_count

    def close(self):
        self._file.close()

    def __repr__(self):
        return """<Frames>
Source: {filename}
Length: {count} frames
Frame Shape: {w}w x {h}h
Pixel Datatype: {dtype}""".format(filename=self.filename,
                                  count=len(self),
                                  h=self.frame_shape[0],
                                  w=self.frame_shape[1],
                                  dtype=self.pixel_type)
